import logging
from typing import Any

from ..const import (
    METERED_NOCT_FALLBACK_TO_INVERTERS,
    URL_PRODUCTION_V1,
    SupportedFeatures,
)
from ..exceptions import ENDPOINT_PROBE_EXCEPTIONS
from ..models.envoy import EnvoyData
from ..models.system_production import EnvoySystemProduction
from .base import EnvoyUpdater

_LOGGER = logging.getLogger(__name__)


class EnvoyApiV1ProductionUpdater(EnvoyUpdater):
    """Class to handle updates for production data."""

    async def probe(
        self, discovered_features: SupportedFeatures
    ) -> SupportedFeatures | None:
        """Probe the Envoy for this updater and return SupportedFeatures."""
        if SupportedFeatures.PRODUCTION in discovered_features:
            # Already discovered from another updater
            return None

        try:
            response: dict[str, Any] = await self._json_probe_request(URL_PRODUCTION_V1)
        except ENDPOINT_PROBE_EXCEPTIONS as e:
            _LOGGER.debug(
                "Production endpoint not found at %s: %s", URL_PRODUCTION_V1, e
            )
            return None

        # Envoy metered without configured CT returns zero's in V1 Production Endpoint
        # sometimes wNow has a value. When Watthours Today, last 7 days and lifetime
        # are all 3 zero is an indication envoy is not reporting summed values in V1 production
        # return None to fallback to inverters section in production endpoint.
        if all(
            value == 0 for key, value in response.items() if key.startswith("wattHours")
        ):
            _LOGGER.debug(
                "Detected broken production endpoint bug at %s: %s",
                URL_PRODUCTION_V1,
                response,
            )
            return None

        # Envoy metered without configured CT in firmware 8.2.8.2.4264 and newer returns
        # stalled values in V1 Production Endpoint. When /info is_meter is set, no configured
        # CT are found and firmware is 8.2.4264 or newer, it is an indication envoy is not reporting proper
        # summed values in V1 production. Return None to fallback to inverters section in production endpoint.
        if (
            self._common_properties.imeter_info
            and self._envoy_version >= METERED_NOCT_FALLBACK_TO_INVERTERS
            and not self._common_properties.ct_meter_count
        ):
            _LOGGER.debug(
                "Detected metered Envoy without connected CT will fallback to using inverter values from /production. fw %s",
                self._envoy_version,
            )
            return None

        self._supported_features |= SupportedFeatures.PRODUCTION
        return self._supported_features

    async def update(self, envoy_data: EnvoyData) -> None:
        """Update the Envoy for this updater."""
        production_data = await self._json_request(URL_PRODUCTION_V1)
        envoy_data.raw[URL_PRODUCTION_V1] = production_data
        envoy_data.system_production = EnvoySystemProduction.from_v1_api(
            production_data
        )



"""Test envoy metered with enabled and disabled CT"""

import logging
from typing import Any

import aiohttp
import jsonpath
import pytest
from aioresponses import aioresponses
from syrupy.assertion import SnapshotAssertion

from pyenphase import register_updater
from pyenphase.const import (
    PHASENAMES,
    SupportedFeatures,
)
from pyenphase.envoy import UPDATERS
from pyenphase.models.meters import (
    CtMeterData,
    CtType,
    EnvoyMeterData,
    EnvoyPhaseMode,
)
from pyenphase.models.system_consumption import EnvoySystemConsumption
from pyenphase.models.system_production import EnvoySystemProduction
from pyenphase.updaters.meters import EnvoyMetersUpdater

from .common import (
    get_mock_envoy,
    load_fixture,
    load_json_fixture,
    load_json_list_fixture,
    override_mock,
    prep_envoy,
    start_7_firmware_mock,
    updater_features,
)

# we're testing, ignore some issue reports
# pyright: reportPrivateUsage=false

LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_pr111_with_7_3_466_metered_disabled_cts(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test envoy metered with disabled ct to report from production inverters PR111."""
    version = "7.3.466_metered_disabled_cts"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert envoy._supported_features & SupportedFeatures.PRODUCTION
    assert envoy._supported_features & SupportedFeatures.INVERTERS
    assert envoy._supported_features & SupportedFeatures.PRODUCTION
    assert updater_features(envoy._updaters) == {
        "EnvoyProductionJsonFallbackUpdater": SupportedFeatures.PRODUCTION,
        "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
        "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
    }
    assert envoy.part_number == "800-00654-r08"

    assert not data.system_consumption
    assert data.system_production is not None
    assert data.system_production.watts_now == 751
    assert data.system_production.watt_hours_today == 4425
    assert data.system_production.watt_hours_last_7_days == 111093
    assert data.system_production.watt_hours_lifetime == 702919


@pytest.mark.asyncio
async def test_pr111_with_7_6_175_with_cts(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test envoy metered with ct to report from production eim PR111."""
    version = "7.6.175_with_cts"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION
    assert envoy._supported_features & SupportedFeatures.NET_CONSUMPTION
    assert envoy._supported_features & SupportedFeatures.PRODUCTION
    assert envoy._supported_features & SupportedFeatures.INVERTERS
    assert envoy._supported_features & SupportedFeatures.METERING
    assert envoy._supported_features & SupportedFeatures.INVERTERS
    assert envoy._supported_features & SupportedFeatures.CTMETERS
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
        "EnvoyProductionJsonUpdater": SupportedFeatures.METERING
        | SupportedFeatures.TOTAL_CONSUMPTION
        | SupportedFeatures.NET_CONSUMPTION
        | SupportedFeatures.PRODUCTION,
        "EnvoyMetersUpdater": SupportedFeatures.CTMETERS,
        "EnvoyTariffUpdater": SupportedFeatures.TARIFF,
    }

    assert envoy.part_number == "800-00654-r08"

    assert data.system_consumption
    assert data.system_production is not None
    assert data.system_production.watts_now == 488
    assert data.system_production.watt_hours_today == 4425
    assert data.system_production.watt_hours_last_7_days == 111093
    assert data.system_production.watt_hours_lifetime == 3183793
    assert (
        envoy.envoy_model
        == "Envoy, phases: 1, phase mode: three, production CT, net-consumption CT"
    )


@pytest.mark.asyncio
async def test_pr111_with_7_6_175_standard(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test envoy metered with ct to report from production eim PR111."""
    version = "7.6.175_standard"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert envoy._supported_features & SupportedFeatures.PRODUCTION
    assert envoy._supported_features & SupportedFeatures.INVERTERS
    assert updater_features(envoy._updaters) == {
        "EnvoyApiV1ProductionInvertersUpdater": SupportedFeatures.INVERTERS,
        "EnvoyApiV1ProductionUpdater": SupportedFeatures.PRODUCTION,
    }

    assert envoy.part_number == "800-00656-r06"

    assert not data.system_consumption
    assert data.system_production is not None
    assert data.system_production.watts_now == 5740
    assert data.system_production.watt_hours_today == 36462
    assert data.system_production.watt_hours_last_7_days == 189712
    assert data.system_production.watt_hours_lifetime == 6139406
    assert envoy.envoy_model == "Envoy"


@pytest.mark.asyncio
async def test_ct_data_structures_with_7_3_466_with_cts_3phase(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test meters model using envoy metered CT with multiple phases"""
    # start with regular data first
    version = "7.3.466_with_cts_3phase"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # details of this test is done elsewhere already, just check data is returned
    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None

    # Preserve the original updaters
    original_updaters = UPDATERS.copy()
    try:
        # Test prior similar updater active
        remove_2nd_metersupdater = register_updater(EnvoyMetersUpdater)
        await envoy.probe()
        remove_2nd_metersupdater()

        # load mock data for meters and their readings
        meters_status = await load_json_list_fixture(version, "ivp_meters")
        meters_readings = await load_json_list_fixture(version, "ivp_meters_readings")

        meter_status: CtMeterData = {
            "eid": meters_status[0]["eid"],
            "state": meters_status[0]["state"],
            "measurementType": meters_status[0]["measurementType"],
            "phaseMode": meters_status[0]["phaseMode"],
            "phaseCount": meters_status[0]["phaseCount"],
            "meteringStatus": meters_status[0]["meteringStatus"],
            "statusFlags": meters_status[0]["statusFlags"],
        }

        # test meters.from_api method
        ct_data: EnvoyMeterData = EnvoyMeterData.from_api(
            meters_readings[0],
            meter_status,
        )
        assert str(ct_data.eid) == "704643328"
        assert ct_data.measurement_type == "production"

        # test meters.from_phase method
        ct_phase_data: EnvoyMeterData | None = EnvoyMeterData.from_phase(
            meters_readings[0], meter_status, 0
        )
        assert ct_phase_data is not None
        assert str(ct_phase_data.eid) == "1778385169"
        assert ct_phase_data.measurement_type == "production"
        assert ct_phase_data.energy_delivered == 3183794

        assert (
            envoy.envoy_model
            == "Envoy, phases: 3, phase mode: three, production CT, net-consumption CT"
        )

        # test exception handling by specifying non-existing phase
        ct_no_phase_data = EnvoyMeterData.from_phase(
            meters_readings[0], meter_status, 3
        )
        assert ct_no_phase_data is None

        # test exception handling for missing phase data, remove phase data from mock data
        del meters_readings[0]["channels"]
        ct_no_phase_data = EnvoyMeterData.from_phase(
            meters_readings[0], meter_status, 0
        )
        assert ct_no_phase_data is None

        # test exception handling for phase data in production using wrong phase
        production_data = data.raw["/production.json?details=1"]
        production_no_phase_data = EnvoySystemProduction.from_production_phase(
            production_data, 3
        )
        assert production_no_phase_data is None

        # test exception handling for phase data if key is missing
        del production_data["production"][1]["type"]
        with pytest.raises(ValueError):
            EnvoySystemProduction.from_production_phase(production_data, 0)

        # test exception handling for phase data in consumption using wrong phase
        consumption_data = data.raw["/production.json?details=1"]
        consumption_no_phase_data = EnvoySystemConsumption.from_production_phase(
            consumption_data, 3
        )
        assert consumption_no_phase_data is None

        # test handling missing phases when expected in ct readings
        meters_status = await load_json_list_fixture(version, "ivp_meters")
        meters_readings = await load_json_list_fixture(version, "ivp_meters_readings")

        # remove phase data from CT readings
        del meters_readings[0]["channels"]
        del meters_readings[1]["channels"]

        override_mock(
            mock_aioresponse,
            "get",
            "https://127.0.0.1/ivp/meters",
            status=200,
            payload=meters_status,
            repeat=True,
        )
        override_mock(
            mock_aioresponse,
            "get",
            "https://127.0.0.1/ivp/meters/readings",
            status=200,
            payload=meters_readings,
            repeat=True,
        )

        await envoy.update()
        data = envoy.data
        assert data is not None
        # should not have phase data after removing phase data from source
        assert data.ctmeters_phases == {}
    finally:
        # Restore the original updaters
        UPDATERS.clear()
        for updater in original_updaters:
            register_updater(updater)


@pytest.mark.asyncio
async def test_ct_data_structures_with_7_6_175_with_cts_3phase(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test meters model using envoy metered CT with multiple phases"""
    # start with regular data first
    version = "7.6.175_with_cts_3phase"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # details of this test is done elsewhere already, just check data is returned
    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None

    # Preserve the original updaters
    original_updaters = UPDATERS.copy()
    try:
        # Test prior similar updater active
        remove_2nd_metersupdater = register_updater(EnvoyMetersUpdater)
        await envoy.probe()
        remove_2nd_metersupdater()

        # load mock data for meters and their readings
        meters_status = await load_json_list_fixture(version, "ivp_meters")
        meters_readings = await load_json_list_fixture(version, "ivp_meters_readings")

        meter_status: CtMeterData = {
            "eid": meters_status[0]["eid"],
            "state": meters_status[0]["state"],
            "measurementType": meters_status[0]["measurementType"],
            "phaseMode": meters_status[0]["phaseMode"],
            "phaseCount": meters_status[0]["phaseCount"],
            "meteringStatus": meters_status[0]["meteringStatus"],
            "statusFlags": meters_status[0]["statusFlags"],
        }

        # test meters.from_api method
        ct_data: EnvoyMeterData = EnvoyMeterData.from_api(
            meters_readings[0],
            meter_status,
        )
        assert str(ct_data.eid) == "704643328"
        assert ct_data.measurement_type == "production"

        # test meters.from_phase method
        ct_phase_data: EnvoyMeterData | None = EnvoyMeterData.from_phase(
            meters_readings[0], meter_status, 0
        )
        assert ct_phase_data is not None
        assert str(ct_phase_data.eid) == "1778385169"
        assert ct_phase_data.measurement_type == "production"
        assert ct_phase_data.energy_delivered == 3183794

        assert (
            envoy.envoy_model
            == "Envoy, phases: 3, phase mode: three, production CT, net-consumption CT"
        )

        # test exception handling by specifying non-existing phase
        ct_no_phase_data = EnvoyMeterData.from_phase(
            meters_readings[0], meter_status, 3
        )
        assert ct_no_phase_data is None

        # test exception handling for missing phase data, remove phase data from mock data
        del meters_readings[0]["channels"]
        ct_no_phase_data = EnvoyMeterData.from_phase(
            meters_readings[0], meter_status, 0
        )
        assert ct_no_phase_data is None

        # test exception handling for phase data if key is missing
        production_data = data.raw["/production.json?details=1"]
        del production_data["production"][1]["type"]
        with pytest.raises(ValueError):
            EnvoySystemProduction.from_production_phase(production_data, 0)
    finally:
        # Restore the original updaters
        UPDATERS.clear()
        for updater in original_updaters:
            register_updater(updater)


@pytest.mark.asyncio
async def test_ct_data_structures_with_7_6_175_with_total_cts_3phase(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test meters model using envoy metered without production CT and total-consumption CT with multiple phases"""
    # start with regular data first
    version = "7.6.175_with_cts_3phase"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)
    production_json = await load_json_fixture(version, "production.json")
    # remove production data to test COV consumption ct only
    del production_json["production"]
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/production.json",
        status=200,
        payload=production_json,
        repeat=True,
    )
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/production.json?details=1",
        status=200,
        payload=production_json,
        repeat=True,
    )

    # Force ct consumption meter to total consumption for COV
    ivp_meters = (await load_fixture(version, "ivp_meters")).replace(
        CtType.NET_CONSUMPTION, CtType.TOTAL_CONSUMPTION
    )
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/ivp/meters",
        status=200,
        body=ivp_meters,
        repeat=True,
    )

    # details of this test is done elsewhere already, just check data is returned
    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None

    assert (
        envoy.envoy_model
        == "Envoy, phases: 3, phase mode: three, production CT, total-consumption CT"
    )


@pytest.mark.asyncio
async def test_ct_storage_with_8_2_127_with_3cts_and_battery_split(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test meters model using envoy metered CT with multiple phases"""
    # start with regular data first
    version = "8.2.127_with_3cts_and_battery_split"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # details of this test is done elsewhere already, just check data is returned
    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None

    # load mock data for meters and their readings
    meters_status = await load_json_list_fixture(version, "ivp_meters")
    meters_readings = await load_json_list_fixture(version, "ivp_meters_readings")

    meter_status: CtMeterData = {
        "eid": meters_status[2]["eid"],
        "state": meters_status[2]["state"],
        "measurementType": meters_status[2]["measurementType"],
        "phaseMode": meters_status[2]["phaseMode"],
        "phaseCount": meters_status[2]["phaseCount"],
        "meteringStatus": meters_status[2]["meteringStatus"],
        "statusFlags": meters_status[2]["statusFlags"],
    }

    # test meters.from_api method
    ct_data: EnvoyMeterData = EnvoyMeterData.from_api(
        meters_readings[2],
        meter_status,
    )
    assert str(ct_data.eid) == "704643840"
    assert ct_data.measurement_type == "storage"

    # test meters.from_phase method
    ct_phase_data: EnvoyMeterData | None = EnvoyMeterData.from_phase(
        meters_readings[2], meter_status, 0
    )
    assert ct_phase_data is not None
    assert str(ct_phase_data.eid) == "1778385681"
    assert ct_phase_data.measurement_type == "storage"
    assert ct_phase_data.energy_delivered == 1136860

    assert (
        envoy.envoy_model
        == "Envoy, phases: 2, phase mode: split, production CT, net-consumption CT, storage CT"
    )

    # test exception handling by specifying non-existing phase
    ct_no_phase_data = EnvoyMeterData.from_phase(meters_readings[2], meter_status, 3)
    assert ct_no_phase_data is None

    # test exception handling for missing phase data, remove phase data from mock data
    del meters_readings[2]["channels"]
    ct_no_phase_data = EnvoyMeterData.from_phase(meters_readings[2], meter_status, 0)
    assert ct_no_phase_data is None

    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/ivp/meters",
        status=200,
        payload=meters_status,
        repeat=True,
    )
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/ivp/meters/readings",
        status=200,
        payload=meters_readings,
        repeat=True,
    )

    await envoy.update()
    data = envoy.data
    assert data is not None
    assert data.ctmeter_storage_phases is None


@pytest.mark.asyncio
async def test_ct_storage_data_without_meter_entry_with_8_2_127_with_3cts_and_battery_split(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test meters model with additional meter readings entry not in meters config"""
    # start with regular data first we use this fixture to test issue reported in 8.3.5025
    version = "8.2.127_with_3cts_and_battery_split"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    # fw D8.3.5027 has 3th (zero) entry for Storage CT, even if not configured
    # this caused Indexerror crash. Test if extra data is now handled without crash
    readings_data = await load_json_list_fixture(version, "ivp_meters_readings")
    readings_data.append({"eid": 1023410688, "channels": [{}, {}, {}]})
    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/ivp/meters/readings",
        status=200,
        payload=readings_data,
        repeat=True,
    )

    # details of this test is done elsewhere already, just check data is returned
    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None


@pytest.mark.asyncio
async def test_yet_unknown_ct_with_8_2_127_with_3cts_and_battery_split(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Test meters model with yet unknown ct type in meters config and readings"""
    # start with regular data first we use this fixture to test issue reported in 8.3.5025
    version = "8.2.127_with_3cts_and_battery_split"
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    yet_unknown_ct_type: str = "this_should_work"

    # change last meter type to one not in CtType
    meters_data = await load_json_list_fixture(version, "ivp_meters")
    meter = meters_data[-1]
    assert meter
    meter["measurementType"] = yet_unknown_ct_type
    del meters_data[-1]
    meters_data.append(meter)

    override_mock(
        mock_aioresponse,
        "get",
        "https://127.0.0.1/ivp/meters",
        status=200,
        payload=meters_data,
        repeat=True,
    )

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data is not None

    # verify yet unknown type is now in ct list and has data with this label
    assert yet_unknown_ct_type in envoy.ct_meter_list
    assert data.ctmeters[yet_unknown_ct_type]
    assert data.ctmeters[yet_unknown_ct_type].state == meter["state"]
    assert data.ctmeters[yet_unknown_ct_type].eid == meter["eid"]
    assert envoy.meter_type(yet_unknown_ct_type) == yet_unknown_ct_type

    # last one in original list was storage ct. Should not be there anymore
    assert data.ctmeter_storage is None
    assert CtType.STORAGE not in data.ctmeters
    assert CtType.STORAGE not in data.ctmeters_phases


@pytest.mark.parametrize(
    ("test_properties",),
    [
        pytest.param(
            {
                "ctMeters": 2,
                "phaseCount": 2,
                "phaseMode": EnvoyPhaseMode.SPLIT,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                ],
            },
            id="4.10.35",
        ),
        pytest.param(
            {
                "ctMeters": 1,
                "phaseCount": 2,
                "phaseMode": EnvoyPhaseMode.SPLIT,
                "meter_types": [
                    CtType.PRODUCTION,
                ],
            },
            id="7.3.130_no_consumption",
        ),
        pytest.param(
            {
                "ctMeters": 2,
                "phaseCount": 3,
                "phaseMode": EnvoyPhaseMode.THREE,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                ],
            },
            id="7.3.466_with_cts_3phase",
        ),
        pytest.param(
            {
                "ctMeters": 2,
                "phaseCount": 2,
                "phaseMode": EnvoyPhaseMode.SPLIT,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                ],
            },
            id="7.3.517_system_2",
        ),
        pytest.param(
            {
                "ctMeters": 2,
                "phaseCount": 1,
                "phaseMode": EnvoyPhaseMode.THREE,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                ],
            },
            id="7.6.175_with_cts",
        ),
        pytest.param(
            {
                "ctMeters": 2,
                "phaseCount": 3,
                "phaseMode": EnvoyPhaseMode.THREE,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                ],
            },
            id="7.6.175_with_cts_3phase",
        ),
        pytest.param(
            {
                "ctMeters": 2,
                "phaseCount": 1,
                "phaseMode": EnvoyPhaseMode.THREE,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                ],
            },
            id="7.6.185_with_cts_and_battery_3t",
        ),
        pytest.param(
            {
                "ctMeters": 3,
                "phaseCount": 2,
                "phaseMode": EnvoyPhaseMode.SPLIT,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                    CtType.STORAGE,
                ],
            },
            id="8.2.127_with_3cts_and_battery_split",
        ),
        pytest.param(
            {
                "ctMeters": 2,
                "phaseCount": 2,
                "phaseMode": EnvoyPhaseMode.SPLIT,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                ],
            },
            id="8.2.127_with_generator_running",
        ),
        pytest.param(
            {
                "ctMeters": 3,
                "phaseCount": 2,
                "phaseMode": EnvoyPhaseMode.SPLIT,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                    CtType.STORAGE,
                ],
            },
            id="8.2.4286_with_3cts_and_battery_split",
        ),
        pytest.param(
            {
                "ctMeters": 2,
                "phaseCount": 1,
                "phaseMode": EnvoyPhaseMode.THREE,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                ],
            },
            id="8.2.4345_with_device_data",
        ),
        pytest.param(
            {
                "ctMeters": 4,
                "phaseCount": 2,
                "phaseMode": EnvoyPhaseMode.SPLIT,
                "meter_types": [
                    CtType.NET_CONSUMPTION,
                    CtType.PRODUCTION,
                    CtType.STORAGE,
                    "backfeed",
                ],
            },
            id="8.3.1598_collar",
        ),
    ],
)
@pytest.mark.asyncio
async def test_current_transformers(
    snapshot: SnapshotAssertion,
    caplog: pytest.LogCaptureFixture,
    test_properties: dict[str, Any],
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
    request: pytest.FixtureRequest,
) -> None:
    """Test Current transformer data and properties."""
    caplog.set_level(logging.WARNING)
    start_7_firmware_mock(mock_aioresponse)

    # verify test parameter completeness
    assert len(test_properties) == 4

    # get version and fixture folder from test id
    version: Any = request.node.callspec.id
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)

    # load data
    data = envoy.data
    assert data is not None
    assert data == snapshot

    # verify expected properties
    assert envoy.ct_meter_count == test_properties["ctMeters"]
    assert envoy.phase_count == test_properties["phaseCount"]
    assert envoy.phase_mode == test_properties["phaseMode"]

    # if we have ct meters we should have CTMETERS feature and if no meters not
    assert envoy.ct_meter_count == len(envoy.ct_meter_list)
    assert envoy._supported_features
    has_ctmeters = bool(envoy._supported_features & SupportedFeatures.CTMETERS)
    meter_count_not_zero = bool(envoy.ct_meter_count > 0)
    assert has_ctmeters == meter_count_not_zero

    # test if expected meters were found
    for cttype in test_properties["meter_types"]:
        assert envoy.meter_type(cttype)
    # test for unexpected meters showing up
    for cttype in envoy.ct_meter_list:
        assert cttype in test_properties["meter_types"]

    # are all CT types represented correctly in model description
    for cttype in envoy.ct_meter_list:
        assert (cttype in envoy.envoy_model) == (envoy.meter_type(cttype) is not None)

    # backward compatibility test, verify individual meter types are still found and in model

    # if no xxx meter is reported then xxx_meter_type should not report one and other way around
    # if no xxx meter is reported then it should not show in modelname and other way around
    has_meter = bool(
        (CtType.TOTAL_CONSUMPTION in envoy.ct_meter_list)
        or (CtType.NET_CONSUMPTION in envoy.ct_meter_list)
    )
    meter_type_present = bool(envoy.consumption_meter_type is not None)
    meter_in_model = bool(str(envoy.consumption_meter_type) in envoy.envoy_model)
    assert has_meter == meter_type_present
    assert has_meter == meter_in_model

    has_meter = bool(CtType.PRODUCTION in envoy.ct_meter_list)
    meter_type_present = bool(envoy.production_meter_type is not None)
    meter_in_model = bool(str(envoy.production_meter_type) in envoy.envoy_model)
    assert has_meter == meter_type_present
    assert has_meter == meter_in_model

    has_meter = bool(CtType.STORAGE in envoy.ct_meter_list)
    meter_type_present = bool(envoy.storage_meter_type is not None)
    meter_in_model = bool(str(envoy.storage_meter_type) in envoy.envoy_model)
    assert has_meter == meter_type_present
    assert has_meter == meter_in_model

    # end backward compatibility test

    # verify meter data
    meter_json = await load_json_fixture(version, "ivp_meters")
    meter_data_json = await load_json_fixture(version, "ivp_meters_readings")

    # get all enabled meters
    enabled_meters: list[Any] = jsonpath.findall("[?(@.state=='enabled')]", meter_json)
    # envoy should have same count
    assert envoy.ct_meter_count == len(enabled_meters)

    # validate each meter data
    for meter in enabled_meters:
        meters_data: list[Any] = jsonpath.findall(
            f"[?(@.eid=={meter['eid']})]", meter_data_json
        )
        assert meters_data[0]
        meter_data = meters_data[0]
        cttype = meter["measurementType"]
        ctdata = data.ctmeters[cttype]
        assert ctdata.energy_delivered == round(meter_data["actEnergyDlvd"])
        assert ctdata.energy_received == round(meter_data["actEnergyRcvd"])
        assert ctdata.active_power == round(meter_data["activePower"])
        assert ctdata.voltage == meter_data["voltage"]
        assert ctdata.current == meter_data["current"]
        assert ctdata.frequency == meter_data["freq"]
        assert ctdata.state == meter["state"]
        assert ctdata.metering_status == meter["meteringStatus"]
        assert ctdata.status_flags == meter.get("statusFlags")

        # backward compatibility test
        # specific xxx meter data should match ctmeters[xxx] data
        meter_match = bool(cttype == CtType.PRODUCTION)
        data_match = bool(data.ctmeter_production == data.ctmeters[cttype])
        assert meter_match == data_match

        meter_match = bool(cttype in (CtType.NET_CONSUMPTION, CtType.TOTAL_CONSUMPTION))
        data_match = bool(data.ctmeter_consumption == data.ctmeters[cttype])
        assert meter_match == data_match

        meter_match = bool(cttype == CtType.STORAGE)
        data_match = bool(data.ctmeter_storage == data.ctmeters[cttype])
        assert meter_match == data_match
        # end backward compatibility test

        # test phase data, if phase count is <=1 no phase data should be present
        multiple_phases = bool(envoy.phase_count > 1)
        phase_data_len_equals_count = bool(
            len(data.ctmeters_phases.get(cttype, {})) == envoy.phase_count
        )
        assert multiple_phases == phase_data_len_equals_count
        for i in range(0, envoy.phase_count if envoy.phase_count > 1 else 0):
            phase_data: Any = jsonpath.findall(
                f"[?(@.eid=={meter['eid']})]['channels'][*]", meter_data_json
            )[i]
            assert data.ctmeters_phases[cttype].get(PHASENAMES[i]) is not None
            ctdata_phase = data.ctmeters_phases[cttype][PHASENAMES[i]]
            assert ctdata_phase.energy_delivered == round(phase_data["actEnergyDlvd"])
            assert ctdata_phase.energy_received == round(phase_data["actEnergyRcvd"])
            assert ctdata_phase.active_power == round(phase_data["activePower"])
            assert ctdata_phase.voltage == phase_data["voltage"]
            assert ctdata_phase.frequency == phase_data["freq"]
            assert ctdata_phase.state == meter["state"]
            assert ctdata_phase.metering_status == meter["meteringStatus"]
            assert ctdata_phase.status_flags == meter.get("statusFlags")

            # backward compatibility, verify individual phase data matches dict data
            # specific xxx meter should match ctmeters_phases[xxx] data
            meter_match = bool(cttype == CtType.PRODUCTION)
            data_match = bool(
                data.ctmeter_production_phases == data.ctmeters_phases.get(cttype)
            )
            assert meter_match == data_match

            meter_match = bool(
                cttype in (CtType.NET_CONSUMPTION, CtType.TOTAL_CONSUMPTION)
            )
            data_match = bool(
                data.ctmeter_consumption_phases == data.ctmeters_phases.get(cttype)
            )
            assert meter_match == data_match

            meter_match = bool(cttype == CtType.STORAGE)
            data_match = bool(
                data.ctmeter_storage_phases == data.ctmeters_phases.get(cttype)
            )
            assert meter_match == data_match
            # end compatibility


@pytest.mark.parametrize(
    ("test_properties",),
    [
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="5.0.62",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="7.3.130",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="7.3.466_metered_disabled_cts",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="7.3.517",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="7.3.517_legacy_savings_mode",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="7.6.114_without_cts",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="7.6.175",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="7.6.175_standard",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="7.6.175_total",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="8.1.41",
        ),
        pytest.param(
            {
                "ctMeters": 0,
                "phaseCount": 1,
                "phaseMode": None,
                "meter_types": [],
            },
            id="8.2.4264_metered_noct",
        ),
    ],
)
@pytest.mark.asyncio
async def test_without_current_transformers(
    snapshot: SnapshotAssertion,
    caplog: pytest.LogCaptureFixture,
    test_properties: dict[str, Any],
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
    request: pytest.FixtureRequest,
) -> None:
    """Test Current transformer data when none are installed."""
    caplog.set_level(logging.WARNING)
    start_7_firmware_mock(mock_aioresponse)

    # verify test parameter completeness
    assert len(test_properties) == 4

    # get version and fixture folder from test id
    version: Any = request.node.callspec.id
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)

    # load data
    data = envoy.data
    assert data is not None
    assert data == snapshot

    # verify expected properties
    assert envoy.ct_meter_count == test_properties["ctMeters"]
    assert envoy.phase_count == test_properties["phaseCount"]
    assert envoy.phase_mode == test_properties["phaseMode"]

    # if we have ct meters we should have CTMETERS feature and not if no meters
    assert envoy.ct_meter_count == len(envoy.ct_meter_list)
    assert envoy._supported_features
    has_ctmeters = bool(envoy._supported_features & SupportedFeatures.CTMETERS)
    meter_count_not_zero = bool(envoy.ct_meter_count > 0)
    assert has_ctmeters == meter_count_not_zero

    # backward compatibility test, verify individual meter types are still found and in model

    # if no xxx meter is reported then xxx_meter_type should not report one and other way around
    # if no xxx meter is reported then it should not show in modelname and other way around
    has_meter = bool(
        (CtType.TOTAL_CONSUMPTION in envoy.ct_meter_list)
        or (CtType.NET_CONSUMPTION in envoy.ct_meter_list)
    )
    meter_type_present = bool(envoy.consumption_meter_type is not None)
    meter_in_model = bool(str(envoy.consumption_meter_type) in envoy.envoy_model)
    assert has_meter == meter_type_present
    assert has_meter == meter_in_model

    has_meter = bool(CtType.PRODUCTION in envoy.ct_meter_list)
    meter_type_present = bool(envoy.production_meter_type is not None)
    meter_in_model = bool(str(envoy.production_meter_type) in envoy.envoy_model)
    assert has_meter == meter_type_present
    assert has_meter == meter_in_model

    has_meter = bool(CtType.STORAGE in envoy.ct_meter_list)
    meter_type_present = bool(envoy.storage_meter_type is not None)
    meter_in_model = bool(str(envoy.storage_meter_type) in envoy.envoy_model)
    assert has_meter == meter_type_present
    assert has_meter == meter_in_model

    # end backward compatibility test



"""Test ACB battery data and combined Encharge/ACB"""

import logging
from typing import Any

import aiohttp
import pytest
from aioresponses import aioresponses
from syrupy.assertion import SnapshotAssertion

from pyenphase.envoy import SupportedFeatures
from pyenphase.models.envoy import EnvoyData

from .common import (
    endpoint_path,
    get_mock_envoy,
    load_json_fixture,
    override_mock,
    prep_envoy,
    start_7_firmware_mock,
)

LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_with_4_2_27_firmware(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Verify with 4.2.27 firmware."""
    version = "4.2.27"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)
    data: EnvoyData | None = envoy.data
    assert data is not None
    assert envoy._supported_features is not None
    assert not (envoy._supported_features & SupportedFeatures.ACB)
    assert not (envoy._supported_features & SupportedFeatures.ENCHARGE)
    assert not (envoy._supported_features & SupportedFeatures.ENPOWER)
    assert envoy.acb_count == 0


@pytest.mark.parametrize(
    (
        "version",
        "supported_features",
        "acb_count",
        "battery_aggregate",
        "acb_power",
    ),
    [
        (
            "4.10.35",
            SupportedFeatures.METERING
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            0,
            {},
            {},
        ),
        (
            "7.3.130",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION,
            0,
            {},
            {},
        ),
        (
            "7.3.130_no_consumption",
            SupportedFeatures.METERING
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            0,
            {},
            {},
        ),
        (
            "7.3.517",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            0,
            {},
            {},
        ),
        (
            "7.3.517_legacy_savings_mode",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            0,
            {},
            {},
        ),
        (
            "7.3.517_system_2",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            0,
            {},
            {},
        ),
        (
            "7.6.175_with_cts",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.CTMETERS,
            0,
            {},
            {},
        ),
        (
            "7.6.175_with_cts_3phase",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.THREEPHASE
            | SupportedFeatures.CTMETERS,
            0,
            {},
            {},
        ),
        (
            "7.3.466_with_cts_3phase",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.THREEPHASE
            | SupportedFeatures.CTMETERS,
            0,
            {},
            {},
        ),
        (
            "7.6.185_with_cts_and_battery_3t",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.TARIFF
            | SupportedFeatures.CTMETERS,
            0,
            {},
            {},
        ),
        (
            "8.1.41",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.PRODUCTION,
            0,
            {},
            {},
        ),
        (
            "8.2.127_with_3cts_and_battery_split",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            0,
            {},
            {},
        ),
        (
            "8.2.127_with_generator_running",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS
            | SupportedFeatures.GENERATOR,
            0,
            {},
            {},
        ),
        (
            "8.2.4382_ACB",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.TARIFF
            | SupportedFeatures.CTMETERS
            | SupportedFeatures.ACB,
            3,
            {
                "available_energy": 2820,
                "max_available_capacity": 7220,
                "state_of_charge": 39,
            },
            {
                "power": 260,
                "charge_wh": 930,
                "state_of_charge": 25,
                "state": "discharging",
                "batteries": 3,
            },
        ),
    ],
    ids=[
        "4.10.35",
        "7.3.130",
        "7.3.130_no_consumption",
        "7.3.517",
        "7.3.517_legacy_savings_mode",
        "7.3.517_system_2",
        "7.6.175_with_cts",
        "7.6.175_with_cts_3phase",
        "7.3.466_with_cts_3phase",
        "7.6.185_with_cts_and_battery_3t",
        "8.1.41",
        "8.2.127_with_3cts_and_battery_split",
        "8.2.127_with_generator_running",
        "8.2.4382_ACB",
    ],
)
@pytest.mark.asyncio
async def test_with_7_x_firmware(
    version: str,
    snapshot: SnapshotAssertion,
    supported_features: SupportedFeatures,
    caplog: pytest.LogCaptureFixture,
    acb_count: int,
    battery_aggregate: dict[str, Any],
    acb_power: dict[str, dict[str, Any]],
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """
    Verify with 7.x firmware.

    Test with fixture that have SupportedFeatures.METERING

    """
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    caplog.set_level(logging.DEBUG)

    envoy = await get_mock_envoy(test_client_session)
    full_host = endpoint_path(version, envoy.host)
    data = envoy.data
    assert data
    assert data == snapshot

    assert envoy.acb_count == acb_count

    # verify both have ACB or both don't have it
    assert not (
        (SupportedFeatures.ACB in envoy.supported_features)
        ^ (SupportedFeatures.ACB in supported_features)
    )
    # verify if acb_power andACB feature are both present or not
    assert not (acb_power != {}) ^ (SupportedFeatures.ACB in envoy.supported_features)

    # verify both are empty/None or both have values
    assert not (acb_power != {}) ^ (data.acb_power is not None)
    assert not (battery_aggregate != {}) ^ (data.battery_aggregate is not None)

    # test battery aggregate values
    for key in battery_aggregate:
        assert data.battery_aggregate is not None
        assert battery_aggregate[key] == getattr(data.battery_aggregate, key)

    # test ACB battery values
    for key in acb_power:
        assert data.acb_power is not None
        assert acb_power[key] == getattr(data.acb_power, key)

    # test for code coverage if no storage section is available
    # use fixtures with METERING in supported_features:
    production_data = data.raw["/production.json?details=1"]

    acb_data = production_data["storage"][0]
    assert acb_data["activeCount"] == acb_count

    # test with missing storage section
    prod_json = await load_json_fixture(version, "production.json")
    del prod_json["storage"]
    override_mock(
        mock_aioresponse,
        "get",
        f"{full_host}/production.json?details=1",
        status=200,
        payload=prod_json,
        repeat=True,
    )
    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data
    assert envoy.acb_count == 0

    # test with activeCount  = 1 and missing percentFull key
    prod_json = await load_json_fixture(version, "production.json")
    prod_json["storage"][0].pop("percentFull", None)
    prod_json["storage"][0]["activeCount"] = 1
    override_mock(
        mock_aioresponse,
        "get",
        f"{full_host}/production.json?details=1",
        status=200,
        payload=prod_json,
        repeat=True,
    )
    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data
    assert envoy.acb_count == 0
    assert data.acb_power is None



"""Enphase Envoy class"""

import asyncio
import logging
import time
from collections.abc import Awaitable, Callable
from dataclasses import replace
from functools import cached_property, partial
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, overload

import aiohttp
import orjson
from awesomeversion import AwesomeVersion
from envoy_utils.envoy_utils import EnvoyUtils
from tenacity import (
    AsyncRetrying,
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    stop_after_delay,
    wait_random_exponential,
)

from pyenphase.models.dry_contacts import DryContactStatus
from pyenphase.models.home import EnvoyInterfaceInformation

from .auth import (
    EnvoyAuth,
    EnvoyLegacyAuth,
    EnvoyTokenAuth,
)
from .const import (
    AUTH_TOKEN_MIN_VERSION,
    DEFAULT_MAX_REQUEST_ATTEMPTS,
    DEFAULT_MAX_REQUEST_DELAY,
    ENDPOINT_URL_HOME,
    LOCAL_TIMEOUT,
    MAX_PROBE_REQUEST_ATTEMPTS,
    MAX_PROBE_REQUEST_DELAY,
    URL_DRY_CONTACT_SETTINGS,
    URL_DRY_CONTACT_STATUS,
    URL_GRID_RELAY,
    URL_TARIFF,
    SupportedFeatures,
)
from .exceptions import (
    EnvoyAuthenticationRequired,
    EnvoyCommunicationError,
    EnvoyError,
    EnvoyFeatureNotAvailable,
    EnvoyHTTPStatusError,
    EnvoyPoorDataQuality,
    EnvoyProbeFailed,
)
from .firmware import EnvoyFirmware
from .json import json_loads
from .models.common import CommonProperties
from .models.envoy import EnvoyData
from .models.meters import CtType, EnvoyPhaseMode
from .models.tariff import EnvoyStorageMode
from .ssl import NO_VERIFY_SSL_CONTEXT
from .updaters.api_v1_production import EnvoyApiV1ProductionUpdater
from .updaters.api_v1_production_inverters import EnvoyApiV1ProductionInvertersUpdater
from .updaters.base import EnvoyUpdater
from .updaters.device_data_inverters import EnvoyDeviceDataInvertersUpdater
from .updaters.ensemble import EnvoyEnembleUpdater
from .updaters.generator import EnvoyGeneratorUpdater
from .updaters.meters import EnvoyMetersUpdater
from .updaters.production import (
    EnvoyProductionJsonFallbackUpdater,
    EnvoyProductionJsonUpdater,
    EnvoyProductionUpdater,
)
from .updaters.tariff import EnvoyTariffUpdater

_LOGGER = logging.getLogger(__name__)


DEFAULT_HEADERS = {
    "Accept": "application/json",
}

UPDATERS: list[type["EnvoyUpdater"]] = [
    EnvoyMetersUpdater,
    EnvoyProductionJsonUpdater,
    EnvoyProductionUpdater,
    EnvoyApiV1ProductionUpdater,
    EnvoyProductionJsonFallbackUpdater,
    EnvoyDeviceDataInvertersUpdater,
    EnvoyApiV1ProductionInvertersUpdater,
    EnvoyEnembleUpdater,
    EnvoyTariffUpdater,
    EnvoyGeneratorUpdater,
]  #: Ordered list of standard updaters for Envoy data collection


def register_updater(updater: type[EnvoyUpdater]) -> Callable[[], None]:
    """
    Register an updater in the updaters list.

    Registers a callable updater at the end of the updaters list. Probe
    method of the registered updater will be called during Envoy.probe().

    Use the returned callable to remove the registered updater again.

    After registering or removing an updater, use Envoy.probe() to make
    the change effective.

    :param updater: callable of (sub-) class EnvoyUpdater
    :return: callable to remove the updater from the registration list
    """
    UPDATERS.append(updater)

    def _remove_updater() -> None:
        """Callable to remove a prior registered updater."""
        UPDATERS.remove(updater)

    return _remove_updater


def get_updaters() -> list[type[EnvoyUpdater]]:
    """
    Return list of registered updaters.

    Returns current list of registered updaters. Changes
    to the list will not be in effect until
    Envoy.probe() is used.

    :return: list of callable updaters
    """
    return UPDATERS


class Envoy:
    """Class for communicating with an envoy."""

    def __init__(
        self,
        host: str,
        client: aiohttp.ClientSession | None = None,
        timeout: float | aiohttp.ClientTimeout | None = None,
    ) -> None:
        """
        Class for communicating with an envoy.

        Collects solar production data for all Envoy models as of
        firmware 3.9. Depending on model and installed components
        can collect power/energy consumption, battery charge,
        discharge and settings. Supports communication with both
        pre- and post V7 Envoy firmware.

        .. code-block:: python

            envoy = Envoy(host_ip_or_name)
            await envoy.setup()
            await envoy.authenticate(
                username=username,
                password=password,
                token=token
            )
            await envoy.update()
            # ...
            await envoy.close()

        :param host: Envoy DNS name or IP address
        :param client: aiohttp ClientSession not verifying SSL
            certificates, if not specified one will be created. In
            that case call :py:meth:`Envoy.close` before application
            exit.
        :param timeout: aiohttp ClientTimeout to use, if not specified
            10 sec connection and 45 sec read timeouts will be used
            (:any:`LOCAL_TIMEOUT`).
        """
        # We use our own aiohttp client session so we can disable SSL verification (Envoys use self-signed SSL certs)
        self._timeout = timeout or LOCAL_TIMEOUT
        connector = aiohttp.TCPConnector(ssl=NO_VERIFY_SSL_CONTEXT)
        self._client = client or aiohttp.ClientSession(connector=connector)  # nosec
        self._user_client = client is not None
        self.auth: EnvoyAuth | None = None
        self._host = host
        self._firmware = EnvoyFirmware(self._client, self._host)
        self._supported_features: SupportedFeatures | None = None
        self._updaters: list[EnvoyUpdater] = []
        self._endpoint_cache: dict[str, aiohttp.ClientResponse] = {}
        self.data: EnvoyData | None = None
        self._common_properties: CommonProperties = CommonProperties()
        self._interface_settings: EnvoyInterfaceInformation | None = None
        self._request_max_attempts: int = DEFAULT_MAX_REQUEST_ATTEMPTS
        self._request_max_delay: int = DEFAULT_MAX_REQUEST_DELAY
        self._request_wait_multiplier: float = 2
        self._request_last_attempts: int = 0
        self._request_last_elapsed: float = 0.0
        self._request_last_endpoint: str = ""

    async def setup(self) -> None:
        """
        Initiate Envoy communication by obtaining firmware version.

        Read /info on Envoy, accessible without authentication.
        Instantiates EnvoyFirmware class object. Required to
        decide what authentication to use for sub-sequent Envoy
        communication. Use this method as first step after class instantiation

        Will retry up to :any:`MAX_PROBE_REQUEST_ATTEMPTS` times or
        :any:`MAX_PROBE_REQUEST_DELAY` elapsed at next try, which
        ever comes first.

        :raises EnvoyFirmwareFatalCheckError: if connection or timeout
            failure occurs
        :raises EnvoyFirmwareCheckError: on http errors or any HTTP
            status other then 200
        """
        await self._firmware.setup()
        # force refetch of interface data next time requested
        self._interface_settings = None

    async def close(self) -> None:
        """
        Close or clean anything opened or created on behalf of the caller.

        Should be called when ending application, if:

        - no aiohttp ClientSession was specified for the Envoy:

          - the pyenphase-created ClientSession will be closed.

        - an aiohttp ClientSession was provided by the caller:

          - Envoy will not close the provided session; the caller remains responsible.

        :return: None
        """
        if not self._user_client and not self._client.closed:
            await self._client.close()

    async def authenticate(
        self,
        username: str | None = None,
        password: str | None = None,
        token: str | None = None,
    ) -> None:
        """
        Authenticate to the Envoy based on firmware version.

        If firmware version retrieved in Envoy.setup is < 7 then create DigestAuth using
        passed username and password. Use 'envoy' or 'installer' username and blank password.

        If Firmware is >= 7 create JWT Token based authorization. If token is passed, use
        it for authorization. If no token is passed, username and password should be
        Enlighten Cloud credentials to obtain a token. Validate the token with the local Envoy.

        :param username: Enligthen Cloud username or local Envoy username, defaults to None
        :param password: Enligthen Cloud password or local Envoy password, defaults to None
        :param token: Token to use with authentication, defaults to None
        :raises EnvoyAuthenticationRequired: Authentication failed with the local Envoy,
            provided token is expired or no token could be obtained from Enlighten cloud
            due to error or missing parameters.
        """
        if self._firmware.version < AUTH_TOKEN_MIN_VERSION:
            # Envoy firmware using old envoy/installer authentication
            _LOGGER.debug(
                "FW: %s, Authenticating to Envoy using envoy/installer authentication",
                self._firmware.version,
            )
            full_serial = self._firmware.serial
            if not username or username == "installer":
                username = "installer"
                password = EnvoyUtils.get_password(full_serial, username)
            elif username == "envoy" and not password:
                # The default password for the envoy user is the last 6 digits of the serial number
                assert full_serial is not None, "Serial must be set"  # nosec
                password = full_serial[-6:]

            if username and password:
                self.auth = EnvoyLegacyAuth(self.host, username, password)

        else:
            # Envoy firmware using new token authentication
            _LOGGER.debug(
                "FW: %s, Authenticating to Envoy using token authentication",
                self._firmware.version,
            )
            if token or (username and password):
                # Always pass all the data to the token auth class, even if some of it is None
                # so that we can refresh the token if needed
                self.auth = EnvoyTokenAuth(
                    self.host,
                    cloud_username=username,
                    cloud_password=password,
                    envoy_serial=self._firmware.serial,
                    token=token,
                )

        if not self.auth:
            _LOGGER.error(
                "You must include username/password or token to authenticate to the Envoy."
            )
            raise EnvoyAuthenticationRequired("Could not setup authentication object.")

        await self.auth.setup(self._client)

    @retry(
        retry=retry_if_exception_type(
            (
                aiohttp.ClientError,
                asyncio.TimeoutError,
            )
        ),
        wait=wait_random_exponential(multiplier=2, max=5),
        stop=stop_after_delay(MAX_PROBE_REQUEST_DELAY)
        | stop_after_attempt(MAX_PROBE_REQUEST_ATTEMPTS),
        reraise=True,
    )
    async def probe_request(self, endpoint: str) -> aiohttp.ClientResponse:
        """
        Make a probe request to the Envoy.

        Probe requests are intended for use during initial
        search of available features in the Envoy. For regular
        data retrieval, use the :py:meth:`pyenphase.Envoy.request`
        or :py:meth:`pyenphase.Envoy.update` methods. Sends GET
        request to endpoint on Envoy and returns the response.

        Probe retries on client connection issues or timeouts.
        Will retry up to :any:`MAX_PROBE_REQUEST_ATTEMPTS` times
        or :any:`MAX_PROBE_REQUEST_DELAY` sec elapsed at next try,
        which ever comes first.

        :param endpoint: Envoy Endpoint to access, start with leading /.
        :raises EnvoyAuthenticationRequired: if no prior authentication
            was completed or HTTP status 401 or 404 is returned.
        :return: request response.
        """
        return await self._request(endpoint)

    async def request(
        self,
        endpoint: str,
        data: dict[str, Any] | None = None,
        method: str | None = None,
    ) -> aiohttp.ClientResponse:
        """
        Make a request to the Envoy.

        Send GET or POST request to Envoy. Defaults to GET, specify
        data dictionary to perform a POST. Only specify the endpoint
        path in the Envoy, HTTP type and Envoy address is prepended
        to form full URL based on authentication method.

        Request retries on client connection issues or timeouts.
        By default will retry up to :any:`DEFAULT_MAX_REQUEST_ATTEMPTS` times or
        :any:`DEFAULT_MAX_REQUEST_DELAY` sec elapsed at next try, which
        ever comes first. Adjust these settings using
        :py:meth:`pyenphase.Envoy.set_retry_policy`

        :param endpoint: Envoy Endpoint to access, start with leading /
        :param data: optional data dictionary to send to the Envoy.
            Defaults to none, if none a GET request is issued.
        :param method: HTTP method to use when sending data dictionary,
            if none and data is specified, POST is default.
        :raises EnvoyAuthenticationRequired: if no prior authentication
            was completed or HTTP status 401 or 404 is returned.
        :raises: Any communication errors when retries are exceeded
        :return: request response.
        """
        self._request_last_attempts = 0
        self._request_last_elapsed = 0
        self._request_last_endpoint = endpoint

        async for attempt in AsyncRetrying(
            wait=wait_random_exponential(
                multiplier=self._request_wait_multiplier, max=5
            ),
            stop=stop_after_delay(self._request_max_delay)
            | stop_after_attempt(self._request_max_attempts),
            retry=retry_if_exception_type(
                (
                    aiohttp.ClientError,
                    asyncio.TimeoutError,
                )
            ),
            reraise=True,
            before_sleep=before_sleep_log(_LOGGER, logging.DEBUG),
        ):
            with attempt:
                result = await self._request(endpoint, data, method)
            self._request_last_attempts = attempt.retry_state.attempt_number
            self._request_last_elapsed = attempt.retry_state.seconds_since_start
        return result

    def set_retry_policy(
        self,
        *,
        max_delay: int | None = None,
        max_attempts: int | None = None,
        wait_multiplier: float | None = None,
    ) -> None:
        """
        Set request retry parameters for request retries. Retry
        attempts will continue until either max_attempts have been
        tried or maximum elapsed time in seconds since first try has
        passed at next attempt. Applies to :py:meth:`pyenphase.Envoy.update`
        and :py:meth:`pyenphase.Envoy.request` methods.

        :param max_delay: maximum time elapsed allowed in seconds since first try.
            Optional, if not specified, setting is not changed. Default setting
            for each instantiated Envoy is :any:`DEFAULT_MAX_REQUEST_DELAY`.
        :param max_attempts: maximum retry attempts. Optional, if not specified,
            setting is not changed. Default setting for each instantiated Envoy
            is :any:`DEFAULT_MAX_REQUEST_ATTEMPTS`.
        :param wait_multiplier: multiplier to increase wait time between
            retry attempts. Optional, if not specified, setting is not changed.
            Default setting for each instantiated Envoy
            is 2. Set to zero to disable waits between retries
        """
        if max_attempts is not None:
            self._request_max_attempts = max_attempts
        if max_delay is not None:
            self._request_max_delay = max_delay
        if wait_multiplier is not None:
            self._request_wait_multiplier = wait_multiplier
        _LOGGER.debug(
            "Request retries set to %s attempts and %s seconds maximum elapsed time, wait multiplier %s",
            self._request_max_attempts,
            self._request_max_delay,
            self._request_wait_multiplier,
        )

    @property
    def last_request_statistics(self) -> dict[str, str | int | float | None]:
        """
        Return statistics of last request call.

        Returns retry information on last executed
        :py:meth:`pyenphase.Envoy.request` method. Provides
        information on retry endpoint, attempts and elapsed time.

        :return: {"endpoint": str,"attempt_number": int,"delay_since_first_attempt": float}.
        """
        return {
            "endpoint": self._request_last_endpoint,
            "attempt_number": self._request_last_attempts,
            "delay_since_first_attempt": self._request_last_elapsed,
        }

    async def _request(
        self,
        endpoint: str,
        data: dict[str, Any] | None = None,
        method: str | None = None,
    ) -> aiohttp.ClientResponse:
        """
        Make a request to the Envoy.

        If data is specified use POST or specified method to
        send data dictionary as json string to the endpoint.
        If no data is specified use GET request. Return the response.

        :param endpoint: Envoy Endpoint to access, start with leading /
        :param data: data dictionary to send to the Envoy, defaults to None
        :param method: method to use to send data dictionary,
            POST if none, only used for data send
        :raises EnvoyAuthenticationRequired: if no prior authentication
            was completed or HTTP status 401 or 404 is returned
        :return: request response
        """
        if self.auth is None:
            raise EnvoyAuthenticationRequired(
                "You must authenticate to the Envoy before making requests."
            )

        url = self.auth.get_endpoint_url(endpoint)
        debugon = _LOGGER.isEnabledFor(logging.DEBUG)
        if debugon:
            request_start = time.monotonic()

        # Set up middleware from auth
        middlewares = (self.auth.auth,) if self.auth.auth else None

        # not using redirects to avoid following 301s to error pages on missing
        # end points and lots of extra requests
        if data:
            if debugon:
                _LOGGER.debug(
                    "Sending %s to %s with data %s",
                    method or "POST",
                    url,
                    orjson.dumps(data),
                )
            response = await self._client.request(
                method or "POST",
                url,
                headers={**DEFAULT_HEADERS, **self.auth.headers},
                timeout=self._timeout,
                data=orjson.dumps(data),
                middlewares=middlewares,
                allow_redirects=False,
            )
        else:
            _LOGGER.debug("Requesting %s with timeout %s", url, self._timeout)
            response = await self._client.get(
                url,
                headers={**DEFAULT_HEADERS, **self.auth.headers},
                timeout=self._timeout,
                middlewares=middlewares,
                allow_redirects=False,
            )

        status_code = response.status
        if status_code in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
            content = await response.read()
            _LOGGER.debug(
                "Authentication failed for %s with status %s: %s",
                url,
                status_code,
                content[:500] if content else "No content",
            )
            raise EnvoyAuthenticationRequired(
                f"Authentication failed for {url} with status {status_code}, "
                "please check your username/password or token."
            )

        # show all responses centrally when in debug
        if debugon:
            request_end = time.monotonic()
            content_type = response.headers.get("content-type")
            _LOGGER.debug(
                "Request reply in %s sec from %s status %s: %s %s",
                round(request_end - request_start, 1),
                url,
                status_code,
                content_type,
                await response.read(),  # Use the actual content bytes
            )

        return response

    async def interface_settings(self) -> EnvoyInterfaceInformation | None:
        """
        Returns Envoy active interface information.

        Returned data includes interface mac, interface type, software
        build date, configured timezone and DHCP settings

        This data is sourced from the /home endpoint which is a slower
        responding endpoint with some potential overhead on the Envoy.
        For this reason, as well as the fact that the data is static,
        it will only be fetched one time when called first time and
        cached internally. Subsequent calls will be fulfilled from the cache.
        A call to envoy.setup() will invalidate the cached data and
        result in a one-time read from Envoy upon next call.

        :return: Interface details or None if error

        """
        if not self._interface_settings:
            try:
                home_json = await self._json_request(
                    end_point=ENDPOINT_URL_HOME, data=None
                )
                self._interface_settings = EnvoyInterfaceInformation.from_api(home_json)
            except EnvoyError as exc:
                _LOGGER.debug("Failure getting interface information %s", exc)
        return self._interface_settings

    @property
    def host(self) -> str:
        """Return the Envoy host specified at initialization."""
        return self._host

    @property
    def firmware(self) -> AwesomeVersion:
        """Return the Envoy firmware version as read from the Envoy."""
        return self._firmware.version

    @property
    def part_number(self) -> str | None:
        """Return the Envoy part number as read from the Envoy."""
        return self._firmware.part_number

    @property
    def serial_number(self) -> str | None:
        """Return the Envoy serial number as read from the Envoy."""
        return self._firmware.serial

    @property
    def is_metered(self) -> bool:
        """Return the Envoy imetered info as read from the Envoy."""
        return self._firmware.is_metered

    @property
    def supported_features(self) -> SupportedFeatures:
        """Return the mask of Envoy supported features as established during Probe."""
        assert self._supported_features is not None, "Call setup() first"  # nosec
        return self._supported_features

    @property
    def phase_count(self) -> int:
        """Return the number of configured phases for CT meters as read from the Envoy."""
        assert self._common_properties is not None, "Call setup() first"  # nosec
        return self._common_properties.phase_count

    @property
    def active_phase_count(self) -> int:
        """Return the number of actual reported phases in Envoy production/consumption report."""
        assert self._common_properties is not None, "Call setup() first"  # nosec
        return self._common_properties.active_phase_count

    @property
    def ct_meter_count(self) -> int:
        """Return the number of configured current transformers (CT) as read from the Envoy"""
        assert self._common_properties is not None, "Call setup() first"  # nosec
        return self._common_properties.ct_meter_count

    @overload
    def _meter_type(self, *types: CtType) -> CtType | None: ...

    @overload
    def _meter_type(self, *types: str) -> str | None: ...

    def _meter_type(self, *types: CtType | str) -> CtType | str | None:
        """Return if ct meter type is installed."""
        assert self._common_properties is not None, "Call setup() first"  # nosec
        if not self._common_properties.meter_types:
            return None
        for type in types:
            if type in self._common_properties.meter_types:
                return type
        return None

    @property
    def consumption_meter_type(self) -> CtType | None:
        """
        Return the type of consumption ct meter installed (total
        or net-consumption or None) as read from the Envoy.
        """
        return self._meter_type(CtType.NET_CONSUMPTION, CtType.TOTAL_CONSUMPTION)

    @property
    def production_meter_type(self) -> CtType | None:
        """
        Return the type of production ct meter installed
        (Production or None) as read from the Envoy.
        """
        return self._meter_type(CtType.PRODUCTION)

    @property
    def storage_meter_type(self) -> CtType | None:
        """Return the type of storage ct meter installed (Storage or None) as read from the Envoy."""
        return self._meter_type(CtType.STORAGE)

    @overload
    def meter_type(self, meter_type: CtType) -> CtType | None: ...

    @overload
    def meter_type(self, meter_type: str) -> str | None: ...

    def meter_type(self, meter_type: CtType | str) -> CtType | str | None:
        """Return the type of ct meter installed as read from the Envoy."""
        return self._meter_type(meter_type)

    @property
    def ct_meter_list(self) -> list[CtType | str]:
        """Return the list of configured current transformers (CT) as read from the Envoy"""
        assert self._common_properties is not None, "Call setup() first"  # nosec
        return self._common_properties.meter_types

    @property
    def phase_mode(self) -> EnvoyPhaseMode | None:
        """Return the phase mode configured for the CT meters (single, split or three) as read from the Envoy."""
        assert self._common_properties is not None, "Call setup() first"  # nosec
        return self._common_properties.phase_mode

    @property
    def acb_count(self) -> int:
        """Return the number of reported ACB batteries in Production storage report section."""
        assert self._common_properties is not None, "Call setup() first"  # nosec
        return self._common_properties.acb_batteries_reported

    @cached_property
    def envoy_model(self) -> str:
        """
        Return Envoy model description.

        Describes the Envoy model based on properties found.

        - if 2 or more phases found or at least 1 ct is found:
        - -  phase count
        - -  phase mode
        - if CT found, type of CTs

        Example: "Envoy, phases: 2, phase mode: split, net-consumption CT, production CT"

        :return: String describing the Envoy model and features.
        """
        model = "Envoy"

        # if CT and more then 1 phase add phase count to model
        ct_count = self.ct_meter_count
        phase_count = self.phase_count
        if phase_count > 1 or ct_count > 0:
            model = f"{model}, phases: {phase_count}"

            # Add phase mode to model
            phase_mode = self.phase_mode
            model = f"{model}, phase mode: {phase_mode}"

        for ct_type in self.ct_meter_list:
            model = f"{model}, {ct_type} CT"

        return model

    async def _make_cached_request(
        self,
        request_func: Callable[[str], Awaitable[aiohttp.ClientResponse]],
        endpoint: str,
    ) -> aiohttp.ClientResponse:
        """Make a cached request."""
        if cached_response := self._endpoint_cache.get(endpoint):
            return cached_response
        response = await request_func(endpoint)
        if response.status == 200:
            self._endpoint_cache[endpoint] = response
        return response

    async def probe(self) -> None:
        """
        Probe for Envoy model and supported features.

        For each updater in the list of updaters returned by get_updaters,
        execute the probe() method. Build and store a list of updaters to use,
        containing updaters for which the probe() method does return at least 1
        supported feature. Store the map of all returned supported features.

        An updaters probe method should determine if the data for the specific
        updater scope is available or not. If so, the updaters update() method will
        be used during data collection.

        Probe should be used only once, after setup and authorization at the start of
        the communication. The update() method will call probe if not done prior.

        :raises EnvoyProbeFailed: if no solar production data can be found on the Envoy.
            Solar production data is available in all Envoy models.
        """
        supported_features = SupportedFeatures(0)
        updaters: list[EnvoyUpdater] = []
        version = self._firmware.version
        metered = self.is_metered
        self._endpoint_cache.clear()
        cached_probe = partial(self._make_cached_request, self.probe_request)
        cached_request = partial(self._make_cached_request, self.request)
        self._common_properties.reset_probe_properties(is_metered=metered)

        for updater in get_updaters():
            klass = updater(
                version, cached_probe, cached_request, self._common_properties
            )
            if updater_features := await klass.probe(supported_features):
                supported_features |= updater_features
                updaters.append(klass)

        if not supported_features & SupportedFeatures.PRODUCTION:
            raise EnvoyProbeFailed("Unable to determine production endpoint")

        self._updaters = updaters
        self._supported_features = supported_features

    def _validate_update(self, data: EnvoyData) -> None:
        """
        Perform some overall validation checks and raise for issues.

        Envoy data returned can be impacted by the state of the Envoy. This
        validates searches for known cases and rules the data as invalid.
        INtention is to avoid false data reported and have it rather signalled
        as poor quality.

        - Envoy firmware v3 starts communicating before all data is established
            and sends zero values for solar production during first minutes
            after startup. Signal poor data quality as long as all values are 0.

        :param data: Envoy data
        :raises EnvoyPoorDataQuality: data was deemed invalid based on specific quality tests
        """
        if self._firmware.version.major == "3" and data.system_production:
            # FW R3.x will return status 200 with all zeros right after startup
            # where never versions return status 503 to signal not ready yet
            # raise error to avoid inserting zero's in historical series.
            production = data.system_production
            if (
                production.watts_now
                + production.watt_hours_today
                + production.watt_hours_last_7_days
                + production.watt_hours_lifetime
            ) == 0:
                raise EnvoyPoorDataQuality(
                    "Data rejected on rule: "
                    f"FW 3.x production all zero at startup ({self._firmware.version})."
                )

    async def update(self) -> EnvoyData:
        """
        Read data from Envoy.

        For each updater in the list of established updaters during probe(),
        execute the update() method to collect current data from the Envoy.
        If probe was never executed, use probe method first.

        An updaters update() method should obtain the data for the specific
        updater scope and save to the Envoy data set.

        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :return: Collected Envoy data
        """
        # Some of the updaters user the same endpoint
        # so we cache the 200 responses for each update
        # cycle to not make duplicate requests
        self._endpoint_cache.clear()
        if not self._supported_features:
            await self.probe()

        data = EnvoyData()
        for updater in self._updaters:
            try:
                await updater.update(data)
            except aiohttp.ClientError as err:  # noqa: PERF203
                raise EnvoyCommunicationError(f"aiohttp ClientError {err!s}") from err
            except asyncio.TimeoutError as err:
                raise EnvoyCommunicationError(f"Timeout {err!s}") from err

        self._validate_update(data)
        self.data = data
        return data

    async def _json_request(
        self, end_point: str, data: dict[str, Any] | None, method: str | None = None
    ) -> Any:
        """
        Make a request to the Envoy and return the JSON response.

        Uses _request() to obtain response and returns response content
        as formatted JSON.

        :param endpoint: Envoy Endpoint to access, start with leading /
        :param data: data dictionary to send to the Envoy, defaults to None
        :param method: method to use to send data dictionary,
            POST if none, only used for data send
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :return: response content as JSON
        """
        try:
            response = await self._request(end_point, data, method)
        except aiohttp.ClientError as err:
            _LOGGER.debug("Request to %s failed with ClientError: %s", end_point, err)
            raise EnvoyCommunicationError(f"aiohttp ClientError {err!s}") from err
        except asyncio.TimeoutError as err:
            _LOGGER.debug("Request to %s timed out: %s", end_point, err)
            raise EnvoyCommunicationError(f"Timeout {err!s}") from err
        if not (200 <= response.status < 300):
            content = await response.read()
            _LOGGER.debug(
                "Request to %s failed with status %s: %s",
                end_point,
                response.status,
                content[:500] if content else "No content",
            )
            raise EnvoyHTTPStatusError(response.status, str(response.url))

        try:
            return json_loads(end_point, await response.read())
        except orjson.JSONDecodeError as err:
            raise EnvoyCommunicationError(
                f"Invalid JSON response from {end_point}: {err}"
            ) from err

    async def go_on_grid(self) -> dict[str, Any]:
        """
        Make a request to the Envoy to go on grid.

        POST {"mains_admin_state": "closed"} to /ivp/ensemble/relay directing
        to connect to the grid. Requires ENPOWER installed.

        :raises EnvoyFeatureNotAvailable: If ENPOWER feature is not available in Envoy
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :return: JSON returned by Envoy
        """
        if not self.supported_features & SupportedFeatures.ENPOWER:
            raise EnvoyFeatureNotAvailable(
                "This feature is not available on this Envoy."
            )
        return await self._json_request(URL_GRID_RELAY, {"mains_admin_state": "closed"})

    async def go_off_grid(self) -> dict[str, Any]:
        """
        Make a request to the Envoy to go off grid.

        POST {"mains_admin_state": "open"} to /ivp/ensemble/relay directing
        to disconnect from the grid. Requires ENPOWER installed.

        :raises EnvoyFeatureNotAvailable: If ENPOWER feature is not available in Envoy
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :return: JSON returned by Envoy
        """
        if not self.supported_features & SupportedFeatures.ENPOWER:
            raise EnvoyFeatureNotAvailable(
                "This feature is not available on this Envoy."
            )
        return await self._json_request(URL_GRID_RELAY, {"mains_admin_state": "open"})

    async def update_dry_contact(self, new_data: dict[str, Any]) -> dict[str, Any]:
        """
        Update settings for an Enpower dry contact relay.

        POST updated dry contact settings to /ivp/ss/dry_contact_settings
        in the Envoy. New_data dict can contain one or more of below items
        to set. The key/value for "id" is required to identify the relay.
        Only include key/values to change.

        .. code-block:: json

            {
                "id": "<relay-id>",
                "grid_action": "value",
                "micro_grid_action": "value",
                "gen_action": "value",
                "override": "value",
                "load_name": "value",
                "mode": "value",
                "soc_low": "value",
                "soc_high": "value",
            },

        Settings specified in the data dict are updated in the
        internally stored dry_contact_settings and send as a whole
        to update the Envoy.

        :param new_data: dict of settings to change, "id" key/value required
        :raises EnvoyFeatureNotAvailable: If ENPOWER feature is not available in Envoy
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :raises ValueError: If update was attempted before first data was requested from Envoy
        :raises ValueError: If no "id" key is present in data dict to send.
        :return: dry_contact_settings JSON returned by Envoy
        """
        # All settings for the relay must be sent in the POST or it may crash the Envoy

        if not self.supported_features & SupportedFeatures.ENPOWER:
            raise EnvoyFeatureNotAvailable(
                "This feature is not available on this Envoy."
            )

        if not (id_ := new_data.get("id")):
            raise ValueError("You must specify the dry contact ID in the data object.")
        # Get the current settings for the relay from EnvoyData and merge with the new settings
        if not (current_data := self.data):
            raise ValueError(
                "Tried to set dry contact settings before the Envoy was queried."
            )
        current_model = current_data.dry_contact_settings[id_]
        new_model = replace(current_model, **new_data)

        return await self._json_request(
            URL_DRY_CONTACT_SETTINGS, {"dry_contacts": new_model.to_api()}
        )

    async def open_dry_contact(self, id: str) -> dict[str, Any]:
        """
        Open a dry contact relay.

        POST {"dry_contacts": {"id": id, "status": "open"}} to Envoy to
        open dry contact with specified id. Upon successful POST,
        update dry contact status in internal data as Envoy needs some time
        to implement the change and have status updated.

        :param id: relay id of dry contact relay to open
        :raises EnvoyFeatureNotAvailable: If ENPOWER feature is not available in Envoy
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :return: JSON response of Envoy
        """
        if not self.supported_features & SupportedFeatures.ENPOWER:
            raise EnvoyFeatureNotAvailable(
                "This feature is not available on this Envoy."
            )

        result = await self._json_request(
            URL_DRY_CONTACT_STATUS, {"dry_contacts": {"id": id, "status": "open"}}
        )
        # The Envoy takes a few seconds before it will reflect the new state of the relay
        # so we preemptively update it
        if TYPE_CHECKING:
            assert self.data is not None  # nosec
        self.data.dry_contact_status[id].status = DryContactStatus.OPEN
        return result

    async def close_dry_contact(self, id: str) -> dict[str, Any]:
        """
        Close a dry contact relay.

        POST {"dry_contacts": {"id": id, "status": "closed"}} to Envoy to
        close dry contact with specified id. Upon successful POST,
        update dry contact status in internal data as Envoy needs some time
        to implement the change and have status updated.

        :param id: relay id of dry contact relay to open
        :raises EnvoyFeatureNotAvailable: If ENPOWER feature is not available in Envoy
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :return: JSON response of Envoy
        """
        if not self.supported_features & SupportedFeatures.ENPOWER:
            raise EnvoyFeatureNotAvailable(
                "This feature is not available on this Envoy."
            )

        result = await self._json_request(
            URL_DRY_CONTACT_STATUS, {"dry_contacts": {"id": id, "status": "closed"}}
        )
        # The Envoy takes a few seconds before it will reflect the new state of the relay
        # so we preemptively update it
        if TYPE_CHECKING:
            assert self.data is not None  # nosec
        self.data.dry_contact_status[id].status = DryContactStatus.CLOSED
        return result

    async def enable_charge_from_grid(self) -> dict[str, Any]:
        """
        Enable charge from grid for Encharge batteries.

        Set charge_from_grid true in internal stored
        tariff data and send updated tariff data to Envoy using PUT.
        This will update the charge from grid setting to true
        in the Envoy.

        :raises EnvoyFeatureNotAvailable: If no Encharge or IQ batteries are available
        :raises EnvoyFeatureNotAvailable: If no TARIFF data is available in Envoy
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :raises ValueError: If update was attempted before first data was requested from Envoy
        :return: JSON response of Envoy
        """
        self._verify_tariff_storage_or_raise()
        if TYPE_CHECKING:
            assert self.data is not None  # nosec
            assert self.data.tariff is not None  # nosec
            assert self.data.tariff.storage_settings is not None  # nosec
        self.data.tariff.storage_settings.charge_from_grid = True
        return await self._json_request(
            URL_TARIFF, {"tariff": self.data.tariff.to_api()}, method="PUT"
        )

    async def disable_charge_from_grid(self) -> dict[str, Any]:
        """
        Disable charge from grid for Encharge batteries.

        Set charge_from_grid false in internal stored
        tariff data and send updated tariff data to Envoy using PUT.
        This will update the charge from grid setting to false
        in the Envoy.

        :raises EnvoyFeatureNotAvailable: If no Encharge or IQ batteries are available
        :raises EnvoyFeatureNotAvailable: If no TARIFF data is available in Envoy
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :raises ValueError: If update was attempted before first data was requested from Envoy
        :return: JSON response of Envoy
        """
        self._verify_tariff_storage_or_raise()
        if TYPE_CHECKING:
            assert self.data is not None  # nosec
            assert self.data.tariff is not None  # nosec
            assert self.data.tariff.storage_settings is not None  # nosec
        self.data.tariff.storage_settings.charge_from_grid = False
        return await self._json_request(
            URL_TARIFF, {"tariff": self.data.tariff.to_api()}, method="PUT"
        )

    async def set_storage_mode(
        self,
        mode: EnvoyStorageMode,
        disable_optimized_schedules: bool = False,
    ) -> dict[str, Any]:
        """
        Set the Encharge storage mode.

        Set storage_mode in internal stored tariff data to specified
        mode and send updated tariff data to /admin/lib/tariff in Envoy
        using PUT. This will update the storage mode setting
        in the Envoy.

        On firmware where optimized schedules are supported, set
        ``disable_optimized_schedules=True`` to also set ``opt_schedules``
        to ``False`` alongside the mode change. When ``opt_schedules`` is
        ``True``, writes to ``storage_settings.mode`` are accepted by the
        gateway and return HTTP 200, but are silently ignored by the battery
        controller. See :ref:`battery-mode-overwritten` for details on cloud reconciliation
        behaviour when the gateway has internet access.

        :param mode: storage mode to set
        :param disable_optimized_schedules: set ``opt_schedules`` to ``False``
            alongside the mode change
        :raises EnvoyFeatureNotAvailable: If no Encharge or IQ batteries are available
        :raises EnvoyFeatureNotAvailable: If no TARIFF data is available in Envoy
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :raises ValueError: If update was attempted before first data was requested from Envoy
        :return: JSON response of Envoy
        """
        self._verify_tariff_storage_or_raise()
        if TYPE_CHECKING:
            assert self.data is not None  # nosec
            assert self.data.tariff is not None  # nosec
            assert self.data.tariff.storage_settings is not None  # nosec
        if type(mode) is not EnvoyStorageMode:
            raise TypeError("Mode must be of type EnvoyStorageMode")
        self.data.tariff.storage_settings.mode = mode
        if (
            disable_optimized_schedules
            and self.data.tariff.storage_settings.opt_schedules is not None
        ):
            self.data.tariff.storage_settings.opt_schedules = False
        return await self._json_request(
            URL_TARIFF, {"tariff": self.data.tariff.to_api()}, method="PUT"
        )

    async def set_reserve_soc(self, value: int) -> dict[str, Any]:
        """
        Set the Encharge reserve state of charge.

        Set reserved_soc in internal stored tariff data to specified
        value and send updated tariff data to /admin/lib/tariff in Envoy
        using PUT. This will update the reserve soc setting
        in the Envoy.

        :param value: reserve soc to set
        :raises EnvoyFeatureNotAvailable: If no Encharge or IQ batteries are available
        :raises EnvoyFeatureNotAvailable: If no TARIFF data is available in Envoy
        :raises EnvoyCommunicationError: when aiohttp network or communication error occurs.
        :raises EnvoyHTTPStatusError: when HTTP status is not 2xx.
        :raises ValueError: If update was attempted before first data was requested from Envoy
        :return: JSON response of Envoy
        """
        self._verify_tariff_storage_or_raise()
        if TYPE_CHECKING:
            assert self.data is not None  # nosec
            assert self.data.tariff is not None  # nosec
            assert self.data.tariff.storage_settings is not None  # nosec
        self.data.tariff.storage_settings.reserved_soc = round(float(value), 1)
        return await self._json_request(
            URL_TARIFF, {"tariff": self.data.tariff.to_api()}, method="PUT"
        )

    def _verify_tariff_storage_or_raise(self) -> None:
        """
        Verify Encharge or IQ batteries and tariff data are available in Envoy

        :raises EnvoyFeatureNotAvailable: If no Encharge or IQ batteries are available
        :raises EnvoyFeatureNotAvailable: If no TARIFF data is available in Envoy
        :raises ValueError: If update was attempted before first data was requested from Envoy
        """
        if not self.supported_features & SupportedFeatures.ENCHARGE:
            raise EnvoyFeatureNotAvailable(
                "This feature requires Enphase Encharge or IQ Batteries."
            )
        if not self.supported_features & SupportedFeatures.TARIFF:
            raise EnvoyFeatureNotAvailable(
                "This feature is not available on this Envoy."
            )
        if not self.data:
            raise ValueError("Tried access envoy data before Envoy was queried.")
        if TYPE_CHECKING:
            assert self.data is not None  # nosec
        if not self.data.tariff:
            raise ValueError(
                "Tried to configure charge from grid before the Envoy was queried."
            )
        if TYPE_CHECKING:
            assert self.data.tariff is not None  # nosec
        if not self.data.tariff.storage_settings:
            raise EnvoyFeatureNotAvailable(
                "This feature requires Enphase Encharge or IQ Batteries."
            )



"""Test endpoint for envoy v7 and newer firmware"""

import logging
from typing import Any

import aiohttp
import pytest
from aioresponses import aioresponses
from syrupy.assertion import SnapshotAssertion

from pyenphase.const import PhaseNames
from pyenphase.envoy import SupportedFeatures
from pyenphase.models.envoy import EnvoyData

from .common import get_mock_envoy, prep_envoy, start_7_firmware_mock

LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_with_4_2_27_firmware(
    mock_aioresponse: aioresponses, test_client_session: aiohttp.ClientSession
) -> None:
    """Verify with 4.2.27 firmware."""
    version = "4.2.27"
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    envoy = await get_mock_envoy(test_client_session)
    data: EnvoyData | None = envoy.data
    assert data is not None
    assert envoy._supported_features is not None

    assert not (envoy._supported_features & SupportedFeatures.TOTAL_CONSUMPTION)
    assert not (envoy._supported_features & SupportedFeatures.NET_CONSUMPTION)
    assert not data.system_net_consumption
    assert not data.system_net_consumption_phases


@pytest.mark.parametrize(
    (
        "version",
        "supported_features",
        "phase_count",
        "net_consumption",
        "net_consumption_phases",
    ),
    [
        (
            "5.0.62",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            1,
            {},
            {},
        ),
        (
            "4.10.35",
            SupportedFeatures.METERING
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            2,
            {},
            {},
        ),
        (
            "7.3.130",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION,
            1,
            {},
            {},
        ),
        (
            "7.3.130_no_consumption",
            SupportedFeatures.METERING
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            2,
            {},
            {},
        ),
        (
            "7.3.517",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            1,
            {},
            {},
        ),
        (
            "7.3.517_legacy_savings_mode",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            1,
            {},
            {},
        ),
        (
            "7.3.517_system_2",
            SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            2,
            {
                "watt_hours_lifetime": -2662919,
                "watt_hours_last_7_days": 0,
                "watt_hours_today": 0,
                "watts_now": 48,
            },
            {},
        ),
        (
            "7.3.466_metered_disabled_cts",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            1,
            {},
            {},
        ),
        (
            "7.6.114_without_cts",
            SupportedFeatures.INVERTERS | SupportedFeatures.PRODUCTION,
            1,
            {},
            {},
        ),
        (
            "7.6.175",
            SupportedFeatures.INVERTERS | SupportedFeatures.PRODUCTION,
            1,
            {},
            {},
        ),
        (
            "7.6.175_total",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF,
            1,
            {},
            {},
        ),
        (
            "7.6.175_standard",
            SupportedFeatures.INVERTERS | SupportedFeatures.PRODUCTION,
            1,
            {
                "watt_hours_lifetime": 1961513,
                "watt_hours_last_7_days": 0,
                "watt_hours_today": 0,
                "watts_now": -11,
            },
            {},
        ),
        (
            "7.6.175_with_cts",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.CTMETERS,
            1,
            {},
            {},
        ),
        (
            "7.6.175_with_cts_3phase",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.THREEPHASE
            | SupportedFeatures.CTMETERS,
            3,
            {
                "watt_hours_lifetime": 2886562,
                "watt_hours_last_7_days": 29891,
                "watt_hours_today": 2200,
                "watts_now": 216,
            },
            {
                PhaseNames.PHASE_1: {
                    "watt_hours_lifetime": 1625201,
                    "watt_hours_last_7_days": 0,
                    "watt_hours_today": 0,
                    "watts_now": 91,
                },
                PhaseNames.PHASE_2: {
                    "watt_hours_lifetime": 629892,
                    "watt_hours_last_7_days": 0,
                    "watt_hours_today": 0,
                    "watts_now": 123,
                },
                PhaseNames.PHASE_3: {
                    "watt_hours_lifetime": 631469,
                    "watt_hours_last_7_days": 0,
                    "watt_hours_today": 0,
                    "watts_now": 1,
                },
            },
        ),
        (
            "7.3.466_with_cts_3phase",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.TARIFF
            | SupportedFeatures.THREEPHASE
            | SupportedFeatures.CTMETERS,
            3,
            {
                "watt_hours_lifetime": 2886562,
                "watt_hours_last_7_days": 0,
                "watt_hours_today": 0,
                "watts_now": 216,
            },
            {
                PhaseNames.PHASE_1: {
                    "watt_hours_lifetime": 1625201,
                    "watt_hours_last_7_days": 0,
                    "watt_hours_today": 0,
                    "watts_now": 91,
                },
                PhaseNames.PHASE_2: {
                    "watt_hours_lifetime": 629892,
                    "watt_hours_last_7_days": 0,
                    "watt_hours_today": 0,
                    "watts_now": 123,
                },
                PhaseNames.PHASE_3: {
                    "watt_hours_lifetime": 631469,
                    "watt_hours_last_7_days": 0,
                    "watt_hours_today": 0,
                    "watts_now": 1,
                },
            },
        ),
        (
            "7.6.185_with_cts_and_battery_3t",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.TARIFF
            | SupportedFeatures.CTMETERS,
            1,
            {
                "watt_hours_lifetime": 1279038,
                "watt_hours_last_7_days": 0,
                "watt_hours_today": 0,
                "watts_now": 525,
            },
            {},
        ),
        (
            "8.1.41",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.PRODUCTION,
            1,
            {
                "watt_hours_lifetime": 1279038,
                "watt_hours_last_7_days": 0,
                "watt_hours_today": 0,
                "watts_now": -7812,
            },
            {},
        ),
        (
            "8.2.127_with_3cts_and_battery_split",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            2,
            {
                "watt_hours_lifetime": 4744550,
                "watt_hours_last_7_days": 0,
                "watt_hours_today": 0,
                "watts_now": 129,
            },
            {},
        ),
        (
            "8.2.127_with_generator_running",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS
            | SupportedFeatures.GENERATOR,
            2,
            {
                "watt_hours_lifetime": 7298714,
                "watt_hours_last_7_days": 0,
                "watt_hours_today": 0,
                "watts_now": 240,
            },
            {},
        ),
        (
            "8.2.4345_with_device_data",
            SupportedFeatures.INVERTERS
            | SupportedFeatures.METERING
            | SupportedFeatures.TOTAL_CONSUMPTION
            | SupportedFeatures.NET_CONSUMPTION
            | SupportedFeatures.PRODUCTION
            | SupportedFeatures.ENCHARGE
            | SupportedFeatures.ENPOWER
            | SupportedFeatures.TARIFF
            | SupportedFeatures.DUALPHASE
            | SupportedFeatures.CTMETERS,
            1,
            {},
            {},
        ),
    ],
    ids=[
        "5.0.62",
        "4.10.35",
        "7.3.130",
        "7.3.130_no_consumption",
        "7.3.517",
        "7.3.517_legacy_savings_mode",
        "7.3.517_system_2",
        "7.3.466_metered_disabled_cts",
        "7.6.114_without_cts",
        "7.6.175",
        "7.6.175_total",
        "7.6.175_standard",
        "7.6.175_with_cts",
        "7.6.175_with_cts_3phase",
        "7.3.466_with_cts_3phase",
        "7.6.185_with_cts_and_battery_3t",
        "8.1.41",
        "8.2.127_with_3cts_and_battery_split",
        "8.2.127_with_generator_running",
        "8.2.4345_with_device_data",
    ],
)
@pytest.mark.asyncio
async def test_with_7_x_firmware(
    version: str,
    snapshot: SnapshotAssertion,
    supported_features: SupportedFeatures,
    caplog: pytest.LogCaptureFixture,
    phase_count: int,
    net_consumption: dict[str, Any],
    net_consumption_phases: dict[str, dict[str, Any]],
    mock_aioresponse: aioresponses,
    test_client_session: aiohttp.ClientSession,
) -> None:
    """Verify with 7.x firmware."""
    start_7_firmware_mock(mock_aioresponse)
    await prep_envoy(mock_aioresponse, "127.0.0.1", version)

    caplog.set_level(logging.DEBUG)

    envoy = await get_mock_envoy(test_client_session)
    data = envoy.data
    assert data
    assert data == snapshot

    assert envoy.phase_count == phase_count

    # are all net consumption phases reported
    expected_phases = net_consumption_phases == {}
    actual_phases = data.system_net_consumption_phases is None
    assert not (expected_phases ^ actual_phases)

    reported_phase_count = envoy.active_phase_count
    # are all net consumption phases reported
    expected_phase_count = len(net_consumption_phases)
    assert expected_phase_count == reported_phase_count

    # are all consumption phases reported
    assert (
        envoy.active_phase_count == 0
        if data.system_net_consumption_phases is None
        else len(data.system_net_consumption_phases)
    )
    # Test each consumption phase
    for phase in net_consumption_phases:
        assert data.system_net_consumption_phases
        assert (consdata := data.system_net_consumption_phases[phase])
        modeldata = net_consumption_phases[phase]

        # test each element of the phase data
        assert consdata.watt_hours_lifetime == modeldata["watt_hours_lifetime"]
        assert consdata.watt_hours_last_7_days == modeldata["watt_hours_last_7_days"]
        assert consdata.watt_hours_today == modeldata["watt_hours_today"]
        assert consdata.watts_now == modeldata["watts_now"]



"""Envoy CT Meter updater"""

import logging
from typing import Any

from ..const import (
    ENDPOINT_URL_METERS,
    ENDPOINT_URL_METERS_READINGS,
    PHASENAMES,
    SupportedFeatures,
)
from ..exceptions import ENDPOINT_PROBE_EXCEPTIONS, EnvoyAuthenticationRequired
from ..models.envoy import EnvoyData
from ..models.meters import CtMeterData, CtState, CtType, EnvoyMeterData, EnvoyPhaseMode
from .base import EnvoyUpdater

_LOGGER = logging.getLogger(__name__)


class EnvoyMetersUpdater(EnvoyUpdater):
    """Class to handle updates for Envoy CT Meters."""

    end_point = ENDPOINT_URL_METERS  #: endpoint in envoy to read CT meter configuration
    data_end_point = (
        ENDPOINT_URL_METERS_READINGS  #: endpoint in Envoy to read CT meter data
    )
    meter_types: list[str]  #: CT types found
    phase_mode: EnvoyPhaseMode | None = (
        None  #: Phase mode configured (Single, Dual or Three)
    )
    phase_count: int = 0  #: Number of phases configured and measured in the Envoy
    ct_meters_count: int = (
        0  #: Number of installed current transformers (Envoy metered Only)
    )
    meter_eids: dict[int | str, str]  #: CT identifiers

    def _set_common_properties(self) -> None:
        """Set Envoy common properties we own and control"""
        self._common_properties.phase_count = self.phase_count
        self._common_properties.phase_mode = self.phase_mode
        self._common_properties.meter_types = self.meter_types
        self._common_properties.ct_meter_count = self.ct_meters_count

    async def probe(
        self, discovered_features: SupportedFeatures
    ) -> SupportedFeatures | None:
        """
        Probe the Envoy meter setup and return CT and multiphase details in SupportedFeatures.

        Get CT configuration info from ivp/meters in the Envoy and determine any multi-phase setup.
        Set :any:SupportedFeatures.CTMETERS if CT are found and enabled.
        Set Threephase or Dualphase supported feature if Envoy is in one of these setups.

        Set common properties (phase_count, ct_meter_count, phase_mode, meter_types)
        to default or discovered values. These are owned by this updater.

        :param discovered_features: Features discovered by other updaters for this updater to skip
        :return: features discovered by this updater
        """
        if SupportedFeatures.CTMETERS in discovered_features:
            # Already discovered from another updater
            return None

        # set defaults for common properties we own and will set
        self.phase_count = 1  # Default to 1 phase which is overall numbers only
        self.ct_meters_count = (
            0  # default no CT, only available on Envoy metered if configured
        )
        self.phase_mode = (
            None  # Phase mode only if ct meters are installed and configured
        )
        # track found ct meter measurement types
        self.meter_types = []

        # set the defaults in global common properties in case we exit early
        self._set_common_properties()

        # set local defaults not shared in common properties
        self.meter_eids = {}

        try:
            meters_json: list[CtMeterData] | None = await self._json_probe_request(
                self.end_point
            )
        except ENDPOINT_PROBE_EXCEPTIONS as e:
            _LOGGER.debug("Meters endpoint not found at %s: %s", self.end_point, e)
            return None
        except EnvoyAuthenticationRequired as e:
            # For D3.18.10 (f0855e) systems return 401 even if the user has access
            # to the endpoint so we must skip it.
            _LOGGER.debug(
                "Skipping meters endpoint as user does not have access to %s: %s",
                self.end_point,
                e,
            )
            return None
        else:
            # The endpoint can return valid json on error
            # in the form of {"error": "message"}
            if not meters_json or "error" in meters_json:
                # Non metered Envoy return empty list
                _LOGGER.debug("No CT Meters found")
                return None
        # Set multiphase features so other providers/models can return phase data
        for meter in meters_json:
            if meter["state"] == CtState.ENABLED:
                # remember what meter is installed
                self.meter_types.append(meter_type := meter["measurementType"])
                # save meter identifier for link between /ivp/meters and /ivp/meters/readings
                self.meter_eids[meter["eid"]] = meter_type

                self.ct_meters_count += 1
                self.phase_mode = meter["phaseMode"]
                self.phase_count = max(self.phase_count, meter["phaseCount"])

        # report phase configuration in envoy common property
        self._set_common_properties()

        # report DUAL or THREE PHASE feature for use by next updaters probe
        if self.phase_count > 2:
            self._supported_features |= SupportedFeatures.THREEPHASE
        elif self.phase_count > 1:
            self._supported_features |= SupportedFeatures.DUALPHASE

        # Signal CTMETERS feature back so update will get used if we found ctmeters
        if self.ct_meters_count > 0:
            self._supported_features |= SupportedFeatures.CTMETERS

        return self._supported_features

    async def update(self, envoy_data: EnvoyData) -> None:
        """
        Update the Envoy data from the meters endpoints.

        Get CT configuration from ivp/meters and CT readings from ivp/meters/readings.
        Store EnvoyMeterData in ctmeters for any meters enabled during probe.
        If more than one phase is active, store per-phase data in ctmeters_phases. Match data
        in ivp/meters and ivp/meters/reading using the eid field in both datasets.

        For backward compatibility, ctmeter_production/ctmeter_consumption/ctmeter_storage
        and their phase equivalents are still set to reference the corresponding entries in
        ctmeters[CtType] and ctmeters_phases[CtType].
        :param envoy_data: EnvoyData structure to store data to
        """
        # get the meter status and readings from the envoy
        meters_status: list[CtMeterData] = await self._json_request(self.end_point)
        meters_readings: list[dict[str, Any]] = await self._json_request(
            self.data_end_point
        )

        envoy_data.raw[self.end_point] = meters_status
        envoy_data.raw[self.data_end_point] = meters_readings

        phase_range = self.phase_count if self.phase_count > 1 else 0

        # no longer assume 2 lists are the same order and size. Size differs in fw 8.3.5025
        status_by_eid = {ct["eid"]: ct for ct in meters_status}
        for meter in meters_readings:
            eid = meter["eid"]

            if not (ct_status := status_by_eid.get(eid)):
                # fw 8.3.5025 also has a 3rd entry for storage ct even if not configured
                # and it has all zeros values. Ignore data if eid not in meter status
                continue

            # match meter identifier to one found during probe to identify production or consumption
            if eid in self.meter_eids:
                # if meter was enabled (eid known) store ctmeter data
                envoy_data.ctmeters[meter_type := self.meter_eids[eid]] = (
                    EnvoyMeterData.from_api(meter, ct_status)
                )
                # if more than 1 phase is configured, store ctmeters phase data
                if phase_data := _meter_data_for_phases(phase_range, meter, ct_status):
                    envoy_data.ctmeters_phases[meter_type] = phase_data

                # Next part is for backward compatibility
                # May plan to remove in some future breaking change version
                if meter_type == CtType.PRODUCTION:
                    envoy_data.ctmeter_production = envoy_data.ctmeters[meter_type]
                    if phase_data:
                        envoy_data.ctmeter_production_phases = (
                            envoy_data.ctmeters_phases[meter_type]
                        )
                elif meter_type in (
                    CtType.NET_CONSUMPTION,
                    CtType.TOTAL_CONSUMPTION,
                ):
                    envoy_data.ctmeter_consumption = envoy_data.ctmeters[meter_type]
                    if phase_data:
                        envoy_data.ctmeter_consumption_phases = (
                            envoy_data.ctmeters_phases[meter_type]
                        )
                elif meter_type == CtType.STORAGE:
                    envoy_data.ctmeter_storage = envoy_data.ctmeters[meter_type]
                    if phase_data:
                        envoy_data.ctmeter_storage_phases = envoy_data.ctmeters_phases[
                            meter_type
                        ]
                # End of backward compatibility


def _meter_data_for_phases(
    phase_range: int, meter: dict[str, Any], ct_data: CtMeterData
) -> dict[str, EnvoyMeterData]:
    """Build a dictionary of phase data for multi-phase setups."""
    meter_data_by_phase: dict[str, EnvoyMeterData] = {
        PHASENAMES[phase_idx]: data
        for phase_idx in range(phase_range)
        if (data := EnvoyMeterData.from_phase(meter, ct_data, phase_idx))
    }
    return meter_data_by_phase



import logging

from ..const import URL_TARIFF, SupportedFeatures
from ..exceptions import ENDPOINT_PROBE_EXCEPTIONS, EnvoyAuthenticationRequired
from ..models.envoy import EnvoyData
from ..models.tariff import EnvoyTariff
from .base import EnvoyUpdater

_LOGGER = logging.getLogger(__name__)


class EnvoyTariffUpdater(EnvoyUpdater):
    """Class to handle updates for the Envoy tariff data."""

    async def probe(
        self, discovered_features: SupportedFeatures
    ) -> SupportedFeatures | None:
        try:
            result = await self._json_probe_request(URL_TARIFF)
        except ENDPOINT_PROBE_EXCEPTIONS as e:
            _LOGGER.debug("Tariff endpoint not found: %s", e)
            return None
        except EnvoyAuthenticationRequired as e:
            # For some systems (Firmware: 3.9.36) return 401 for
            # this endpoint even if the user is authenticated.
            _LOGGER.debug(
                "Skipping tariff endpoint as user does not have access to %s: %s",
                URL_TARIFF,
                e,
            )
            return None
        else:
            if not result or "error" in result:
                _LOGGER.debug("No tariff data found")
                return None
            self._supported_features |= SupportedFeatures.TARIFF
        return self._supported_features

    async def update(self, envoy_data: EnvoyData) -> None:
        raw = await self._json_request(URL_TARIFF)
        envoy_data.raw[URL_TARIFF] = raw

        envoy_data.tariff = EnvoyTariff.from_api(raw["tariff"])



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "pyenphase"
copyright = "2023-2025, pyenphase"
author = "pyenphase"
release = "|release|"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
]

# The suffix of source filenames.
source_suffix = [
    ".rst",
    ".md",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Generate internal ancjors for headings
myst_heading_anchors = 4

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



"""Envoy production data updater"""

import logging
from typing import Any

from ..const import PHASENAMES, URL_PRODUCTION, URL_PRODUCTION_JSON, SupportedFeatures
from ..exceptions import ENDPOINT_PROBE_EXCEPTIONS, EnvoyAuthenticationRequired
from ..models.acb import EnvoyACBPower
from ..models.envoy import EnvoyData
from ..models.system_consumption import EnvoySystemConsumption
from ..models.system_production import EnvoySystemProduction
from .base import EnvoyUpdater

_LOGGER = logging.getLogger(__name__)


class EnvoyProductionUpdater(EnvoyUpdater):
    """Class to handle updates for production data."""

    end_point = URL_PRODUCTION
    allow_inverters_fallback = False

    async def probe(
        self, discovered_features: SupportedFeatures
    ) -> SupportedFeatures | None:
        """Probe the Envoy for this endpoint and return SupportedFeatures."""
        discovered_total_consumption = (
            SupportedFeatures.TOTAL_CONSUMPTION in discovered_features
        )
        discovered_net_consumption = (
            SupportedFeatures.NET_CONSUMPTION in discovered_features
        )
        discovered_production = SupportedFeatures.PRODUCTION in discovered_features

        # check if prior updaters reported supporting ACB
        discovered_acb_storage = SupportedFeatures.ACB in discovered_features

        # obtain any registered production endpoints that replied back from the common list
        # when in allow_inverters_fallback mode we can use the first one that worked
        working_endpoints: list[str] = self._common_properties.production_fallback_list
        if (
            discovered_total_consumption
            and discovered_net_consumption
            and discovered_production
        ):
            # Already discovered from another updater
            return None

        # when active allow_inverters_fallback use first successful endpoint registered in the list
        if self.allow_inverters_fallback and working_endpoints:
            self.end_point = working_endpoints[0]

        try:
            production_json: dict[str, Any] = await self._json_probe_request(
                self.end_point
            )
        except ENDPOINT_PROBE_EXCEPTIONS as e:
            _LOGGER.debug("Production endpoint not found at %s: %s", self.end_point, e)
            return None
        except EnvoyAuthenticationRequired as e:
            # For URL_PRODUCTION some systems return 401 even if the user has access
            # to the endpoint. For URL_PRODUCTION_JSON some early non-metered V7 versions
            # return 401 when using aiohttp. Non-metered fallback to v1 production anyway.
            # if there's really no-auth the v1/production will catch it.
            if self.end_point in (URL_PRODUCTION, URL_PRODUCTION_JSON):
                _LOGGER.debug(
                    "Skipping production endpoint as user does"
                    " not have access to %s: %s",
                    self.end_point,
                    e,
                )
                return None
            _LOGGER.debug(  # pragma: no cover
                "Authentication required for %s, re-raising exception: %s",
                self.end_point,
                e,
            )
            raise  # pragma: no cover

        active_phase_count = 0
        phase_count = self._common_properties.phase_count

        # if endpoint is not in the list of successful endpoints yet, add it.
        if (
            self.end_point not in working_endpoints
            and not self.allow_inverters_fallback
        ):
            working_endpoints.append(self.end_point)

        if not discovered_production:
            production: list[dict[str, Any]] | None = production_json.get("production")
            if production:
                for type_ in production:
                    if type_["type"] == "eim" and type_["activeCount"]:
                        self._supported_features |= SupportedFeatures.METERING
                        self._supported_features |= SupportedFeatures.PRODUCTION
                        if lines := type_.get("lines"):
                            active_phase_count = len(lines)
                        break
                    if (
                        self.allow_inverters_fallback
                        and type_["type"] == "inverters"
                        and type_["activeCount"]
                    ):
                        self._supported_features |= SupportedFeatures.PRODUCTION
                        break

        consumption: list[dict[str, Any]] | None = production_json.get("consumption")
        if consumption:
            for meter in consumption:
                # first test activecount >0 before trying measurementType
                if not meter.get("activeCount"):
                    continue
                meter_type = meter["measurementType"]
                if (
                    not discovered_total_consumption
                    and meter_type == "total-consumption"
                ):
                    self._supported_features |= SupportedFeatures.TOTAL_CONSUMPTION
                if not discovered_net_consumption and meter_type == "net-consumption":
                    self._supported_features |= SupportedFeatures.NET_CONSUMPTION
                if lines := meter.get("lines"):
                    active_phase_count = len(lines)

        acb_storage: list[dict[str, Any]] | None = production_json.get("storage")
        # if storage segment is present and activeCount > 0 then signal as detected
        # verify presence of percentFull key to confirm a well configured envoy
        # we've seen activeCount > 0 without percentFull key, ignore ACB in that case
        # only report we support ACB if no prior updater did
        # only report first list element as that's where ACB data is
        if (
            acb_storage
            and not discovered_acb_storage
            and (acb_count := acb_storage[0].get("activeCount"))
            and "percentFull" in acb_storage[0]
        ):
            # signal we can provide ACB data
            self._supported_features |= SupportedFeatures.ACB
            # save acb batt count in common properties so Ensemble can report combined soc
            self._common_properties.acb_batteries_reported = acb_count

        # register the updated fallback endpoints to the common list
        self._common_properties.production_fallback_list = working_endpoints
        self._common_properties.active_phase_count = active_phase_count
        if active_phase_count != phase_count and phase_count > 1:
            _LOGGER.debug(
                "Expected Production report Phase values not available, %s of %s",
                active_phase_count,
                phase_count,
            )
        return self._supported_features

    async def update(self, envoy_data: EnvoyData) -> None:
        """Update the Envoy for this endpoint."""
        production_data = await self._json_request(self.end_point)
        envoy_data.raw[self.end_point] = production_data

        # get phase count from Envoy common features
        phase_count = self._common_properties.phase_count

        if self._supported_features & SupportedFeatures.PRODUCTION:
            envoy_data.system_production = EnvoySystemProduction.from_production(
                production_data
            )
            # get production phase data if more then 1 phase is found
            phase_production: dict[str, EnvoySystemProduction | None] = {}
            for phase in range(phase_count if phase_count > 1 else 0):
                production: EnvoySystemProduction | None = (
                    EnvoySystemProduction.from_production_phase(production_data, phase)
                )
                # exclude None phases that are expected but not actually in production report
                if production:
                    phase_production[PHASENAMES[phase]] = production

            if len(phase_production) > 0:
                envoy_data.system_production_phases = phase_production

        if (
            self._supported_features & SupportedFeatures.NET_CONSUMPTION
            or self._supported_features & SupportedFeatures.TOTAL_CONSUMPTION
        ):
            envoy_data.system_consumption = EnvoySystemConsumption.from_production(
                production_data
            )
            envoy_data.system_net_consumption = EnvoySystemConsumption.from_production(
                production_data, consumption_segment=1
            )

            # get consumption phase data if more then 1 phase is found
            phase_consumption: dict[str, EnvoySystemConsumption | None] = {}
            phase_net_consumption: dict[str, EnvoySystemConsumption | None] = {}
            for phase in range(phase_count if phase_count > 1 else 0):
                consumption: EnvoySystemConsumption | None = (
                    EnvoySystemConsumption.from_production_phase(production_data, phase)
                )
                # exclude None phases that are expected but not actually in production report
                if consumption:
                    phase_consumption[PHASENAMES[phase]] = consumption

                if net_consumption := (
                    EnvoySystemConsumption.from_production_phase(
                        production_data, phase, consumption_segment=1
                    )
                ):
                    phase_net_consumption[PHASENAMES[phase]] = net_consumption

            if len(phase_consumption) > 0:
                envoy_data.system_consumption_phases = phase_consumption

            if len(phase_net_consumption) > 0:
                envoy_data.system_net_consumption_phases = phase_net_consumption

        if self._supported_features & SupportedFeatures.ACB:
            envoy_data.acb_power = EnvoyACBPower.from_production(production_data)


class EnvoyProductionJsonUpdater(EnvoyProductionUpdater):
    """Class to handle updates for production data from the production.json endpoint."""

    end_point = URL_PRODUCTION_JSON


class EnvoyProductionJsonFallbackUpdater(EnvoyProductionJsonUpdater):
    """
    Class to handle updates for production data from the production.json endpoint.

    This class will accept the production endpoint even if activeCount is 0
    """

    allow_inverters_fallback = True
